"""
DGCV: Differential Geometry with Complex Variables

This module provides tools for creating, manipulating, and decomposing vector fields and 
differential forms within the DGCV package. It includes methods for Cartan calculus like 
the exterior derivative and Lie derivative. There are some complex structure specific
functions as well, e.g., assembling holomorphic vector fields from holomorphic coefficients.

Key Functions:

Object Assembly:
    - get_VF(): Retrieves coordinate vector fields for the given coordinate variables.
    - get_DF(): Retrieves coordinate differential forms for the given coordinate variables.
    - assembleFromHolVFC(): Assembles a holomorphic vector field from holomorphic coefficients.
    - assembleFromAntiholVFC(): Assembles an antiholomorphic vector field from antiholomorphic 
    coefficients.
    - assembleFromCompVFC(): Assembles a complex vector field from both holomorphic and 
    antiholomorphic coefficients.

Differential Operators:
    - exteriorDerivative(): Computes the exterior derivative of a differential k-form.
    - interiorProduct(): Computes the interior product of a vector field with a differential k-form.
    - LieDerivative(): Computes the Lie derivative of a differential form or vector field with respect to another vector field.

Decompositions and Basis:
    - decompose(): Decomposes a vector field or differential form as a linear combination of a given basis of vector fields or differential forms.
    - get_coframe(): Constructs a coframe dual to a given list of vector fields.
    - annihilator(): Computes the annihilator (namely sub-bundle in TM/T^*M) of a list of differential forms or vector fields.

Author: David Sykes (https://github.com/YikesItsSykes)

Dependencies:
    - sympy

License:
    MIT License
"""

############## dependencies
import warnings

import sympy as sp

from ._safeguards import create_key, retrieve_passkey
from .combinatorics import chooseOp
from .config import _cached_caller_globals, get_variable_registry
from .dgcv_core import (
    DFClass,
    VF_bracket,
    VFClass,
    _format_complex_coordinates,
    _remove_complex_handling,
    addDF,
    addVF,
    allToReal,
    allToSym,
    changeDFBasis,
    clearVar,
    compressDGCVClass,
    listVar,
    minimalVFDataDict,
    variableProcedure,
)


############## retrieval
def get_VF(*coordinates):
    vr = get_variable_registry()
    VFList = []
    parentS = vr["standard_variable_systems"]
    parentC = vr["complex_variable_systems"]
    parents = parentS | parentC
    for j in coordinates:
        varStr = str(j)
        for parent in parentS | parentC:
            if parent in parentS:
                the_house = parentS[parent]["family_values"]
            else:
                if isinstance(parentC[parent]["family_values"][0], tuple):
                    the_house = sum(parentC[parent]["family_values"], tuple())
                else:
                    the_house = parentC[parent]["family_values"]
            if j in the_house:
                if parents[parent]["differential_system"]:
                    VFList += [parents[parent]["variable_relatives"][varStr]["VFClass"]]
                else:
                    raise TypeError(
                        f"`get_VF` recieved a variable {j} that was not initialized with coordinate vector fields registered in DGCV's variable management framework. Use functions like `createVariables(--,withVF=True)` and `createVariables(--,complex=True)` to initialize variables with VF and DF."
                    )
    return VFList


def get_DF(*coordinates):
    vr = get_variable_registry()
    DFList = []
    parentS = vr["standard_variable_systems"]
    parentC = vr["complex_variable_systems"]
    parents = parentS | parentC
    for j in coordinates:
        varStr = str(j)
        for parent in parentS | parentC:
            if parent in parentS:
                the_house = parentS[parent]["family_values"]
            else:
                if isinstance(parentC[parent]["family_values"][0], tuple):
                    the_house = sum(parentC[parent]["family_values"], tuple())
                else:
                    the_house = parentC[parent]["family_values"]
            if j in the_house:
                if parents[parent]["differential_system"]:
                    DFList += [parents[parent]["variable_relatives"][varStr]["DFClass"]]
                else:
                    raise TypeError(
                        "`get_DF` recieved variables that were not initialized with coordinate vector fields and dual 1-forms registered in DGCV's variable management framework. Use functions like `createVariables(--,withVF=True)` and `createVariables(--,complex=True)` to initialize variables with VF and DF."
                    )
    return DFList


############## complex vector fields


def assembleFromHolVFC(coeffs, holVars):
    """
    Constructs a holomorphic vector field (i.e., VFClass instance) with prescribed coefficients in chosen coordinates.

    The vector field is expressed as a linear combination of the coordinate holomorphic vector fields
    corresponding to the variables in *holVars* with coefficients in *coeffs*.

    Parameters:
    -----------
    coeffs : tuple of sympy expressions
        Coefficients for the vector field in terms of the holomorphic variables.
    holVars : list or tuple
        A list or tuple containing Symbol objects that were initialized as holomorphic variables via DGCV variable creation functions.

    Returns:
    --------
    VFClass
        A holomorphic vector field expressed as a linear combination of coordinate vector fields.

    Raises:
    -------
    ValueError
        If the length of *coeffs* does not match the number of holomorphic variables in *holVars*.
    Exception
        If variables in *holVars* were not initialized as holomorphic variables.

    Example:
    --------
    >>> from DGCV import createVariables, assembleFromHolVFC
    >>> createVariables('z', 'x', 'y', 3)
    >>> print(assembleFromHolVFC((z1 + z2, z3, 0), [z1, z2, z3]))
    (z1/2 + z2/2)*D_x1+z3/2*D_x2+(-I*(z1 + z2)/2)*D_y1-I*z3/2*D_y2
    """
    # Check that length of coefficients matches the variable space
    if len(coeffs) != len(holVars):
        raise ValueError(
            f"The number of coefficients in coeffs must match the number of variables in holVars. Instead given \n coeffs: {coeffs} \n vars: {holVars}."
        )
    vr = get_variable_registry()
    if all(var in vr["conversion_dictionaries"]["holToReal"] for var in holVars):
        pass
    else:
        problem_vars = []
        for var in holVars:
            if var not in vr["conversion_dictionaries"]["symToReal"]:
                problem_vars += [var]
        raise TypeError(
            f"`assembleFromHolVFC` expects the provided varibles for the coordinate space to be holomorphic. The following given vars do not satisfy this: {problem_vars}"
        )
    antiHolVars = tuple([vr['conversion_dictionaries']['conjugation'][var] for var in holVars])
    vars = tuple(holVars)+antiHolVars
    coeffs = tuple(coeffs)+((0,)*len(holVars))

    # Return the resulting vector field as a VFClass instance
    return VFClass(vars, coeffs, "complex")


def assembleFromAntiholVFC(coeffs, holVars):
    """
    Constructs an antiholomorphic vector field (i.e., VFClass instance) with prescribed coefficients in chosen coordinates.

    The vector field is expressed as a linear combination of the coordinate antiholomorphic vector fields
    corresponding to the variables in *arg2* with coefficients in *arg1*.

    Parameters:
    -----------
    arg1 : tuple of sympy expressions
        Coefficients for the vector field in terms of the antiholomorphic variables.
    arg2 : list or tuple
        A list or tuple containing Symbol objects that were initialized as holomorphic variables via DGCV variable creation functions.

    Returns:
    --------
    VFClass
        An antiholomorphic vector field expressed as a linear combination of coordinate vector fields.

    Raises:
    -------
    ValueError
        If the length of *arg1* does not match the number of holomorphic variables in *arg2*.
    Exception
        If variables in *arg2* were not initialized as holomorphic variables.

    Example:
    --------
    >>> from DGCV import createVariables, assembleFromAntiholVFC
    >>> createVariables('z', 'x', 'y', 3)
    >>> print(assembleFromAntiholVFC((z1 + z2, z3, 0), [z1, z2, z3]))
    (z1/2 + z2/2)*D_x1+z3/2*D_x2+(I*(z1 + z2)/2)*D_y1+I*z3/2*D_y2
    """

    # Check that length of coefficients matches the variable space
    if len(coeffs) != len(holVars):
        raise ValueError(
            f"The number of coefficients in coeffs must match the number of variables in holVars. Instead given \n coeffs: {coeffs} \n vars: {holVars}."
        )
    vr = get_variable_registry()
    if all(var in vr["conversion_dictionaries"]["holToReal"] for var in holVars):
        pass
    else:
        problem_vars = []
        for var in holVars:
            if var not in vr["conversion_dictionaries"]["symToReal"]:
                problem_vars += [var]
        raise TypeError(
            f"`assembleFromHolVFC` expects the provided varibles for the coordinate space to be holomorphic. The following given vars do not satisfy this: {problem_vars}"
        )
    antiHolVars = tuple([vr['conversion_dictionaries']['conjugation'][var] for var in holVars])
    vars = tuple(holVars)+antiHolVars
    coeffs = ((0,)*len(holVars)) + tuple(coeffs)

    # Return the resulting vector field as a VFClass instance
    return VFClass(vars, coeffs, "complex")


def assembleFromCompVFC(arg1, arg2, arg3):
    """
    Constructs a complex vector field (i.e., VFClass instance) with prescribed coefficients of the vector field's
    holomorphic and antiholomorphic parts in chosen coordinates.

    The vector field is expressed as a linear combination of the coordinate holomorphic and antiholomorphic vector
    fields corresponding to the variables in *arg3*, with holomorphic part coefficients in *arg1* and antiholomorphic
    part coefficients in *arg2*.

    Parameters:
    -----------
    arg1 : tuple of sympy expressions
        Coefficients for the holomorphic part of the vector field.
    arg2 : tuple of sympy expressions
        Coefficients for the antiholomorphic part of the vector field.
    arg3 : list or tuple
        A list or tuple containing Symbol objects that were initialized as holomorphic variables via DGCV variable creation functions.

    Returns:
    --------
    VFClass
        A complex vector field expressed as a linear combination of holomorphic and antiholomorphic coordinate vector fields.

    Raises:
    -------
    ValueError
        If the length of *arg1* or *arg2* does not match the number of variables in *arg3*.
    Exception
        If variables in *arg3* were not initialized as holomorphic variables.

    Example:
    --------
    >>> from DGCV import createVariables, assembleFromCompVFC
    >>> createVariables('z', 'x', 'y', 3)
    >>> assembleFromCompVFC((z1 + z2, z3, 0), (z1, z2, z3), [z1, z2, z3])
    VFClass instance representing the complex vector field with holomorphic and antiholomorphic parts
    """
    # check that the length of both sets of coefficients matches the variable space
    if len(arg1) != len(arg3) or len(arg2) != len(arg3):
        raise ValueError(
            "The number of coefficients in arg1 and arg2 must match the number of variables in arg3."
        )

    # Construct the holomorphic and antiholomorphic vector fields and add them
    hol_vf = assembleFromHolVFC(arg1, arg3)
    antihol_vf = assembleFromAntiholVFC(arg2, arg3)

    # Return the resulting complex vector field
    return addVF(hol_vf, antihol_vf)


############## differential forms


def makeZeroForm(arg1, varSpace=None, DGCVType=None, default_var_format=None):
    """
    Constructs a 0-form (a differential form of degree 0) based on the provided input.

    If *varSpace* is not provided, the variables will be inferred from the real parts of *arg1*. If *DGCVType* is
    not provided, the function checks if the variables belong to the complex variable system (via the variable registry)
    to determine the differential form type.

    Parameters:
    -----------
    arg1 : sympy expression or numeric value
        The scalar expression or value for the 0-form.
    varSpace : list or tuple, optional
        The variable space associated with the form. If not provided, the real free symbols of *arg1* will be used.
    DGCVType : str, optional
        The differential form type ('complex' or 'standard'). If not provided, the type will be inferred from the
        variable space.

    Returns:
    --------
    DFClass
        A new differential form instance of degree 0.

    Example:
    --------
    >>> from DGCV import makeZeroForm, createVariables
    >>> createVariables('a b')
    >>> createVariables('z', 'x', 'y')
    >>> df1 = makeZeroForm(5)  # constant coeff DF form with trivial varSpace
    >>> df2 = makeZeroForm(a + b)  # standard type DFClass instance representing the 0-form with expression a + b and varSpace (a, b)
    >>> df3 = makeZeroForm(z * BARz)  # complex type DFClass instance representing the 0-form with expression x**2 + y**2 and varSpace (x, y)
    """
    variable_registry = get_variable_registry()

    # Determine the variable space
    if varSpace is None:
        if isinstance(arg1, (int, float)):
            varLoc = tuple()
        elif isinstance(arg1, sp.Expr):
            if DGCVType == "complex":
                if default_var_format == "real":
                    varLoc = tuple(
                        [
                            j
                            for j in allToReal(arg1).free_symbols
                            if j
                            in variable_registry["conversion_dictionaries"]["realToSym"]
                        ]
                    )
                else:
                    default_var_format = "complex"
                    varLoc = tuple(
                        [
                            j
                            for j in allToSym(arg1).free_symbols
                            if j
                            in variable_registry["conversion_dictionaries"]["symToReal"]
                        ]
                    )
            else:
                varLoc = tuple(allToReal(arg1).free_symbols)
    else:
        if DGCVType == "complex":
            if (
                all(
                    var in variable_registry["conversion_dictionaries"]["realToSym"]
                    for var in varSpace
                )
                and default_var_format != "complex"
            ):
                varLoc = tuple(varSpace)
            elif (
                all(
                    var in variable_registry["conversion_dictionaries"]["symToReal"]
                    for var in varSpace
                )
                and default_var_format != "real"
            ):
                varLoc = tuple(varSpace)
            else:
                raise TypeError(
                    "`makeZeroForm` received either all real coordinates for `varSpace` while `default_var_format='complex'` was set, all complex coordinates for `varSpace` while `default_var_format='real'` was set, or was given a set for `varspace` with a mixture of real and complex coordinates. None of these data types are supported in `makeZeroForm`. Tip: Specifying varSpace is optional, so if trying to create a complex type 0-form with the keyword `default_var_format` specified, consider not specifying the varSpace variables."
                )
        else:
            varLoc = tuple(varSpace)

    # Determine the differential form type
    if DGCVType is None:
        if all(
            var in variable_registry["conversion_dictionaries"]["realToSym"]
            for var in varLoc
        ):
            typeLoc = "complex"
        elif all(
            var in variable_registry["conversion_dictionaries"]["symToReal"]
            for var in varLoc
        ):
            typeLoc = "complex"
        else:
            typeLoc = "standard"  # default to standard unless found in complex variable system
    else:
        typeLoc = DGCVType

    # Return the 0-form as a DFClass instance
    return DFClass(varLoc, {tuple(): arg1}, 0, DGCVType=typeLoc)


def exteriorDerivative(arg1, forceComplexType=None):
    """
    Computes the exterior derivative of a differential form (i.e., DFClass instance).

    If *arg1* is not already a differential form, it is converted into a 0-form via makeZeroForm. The function
    supports handling both standard and complex forms, with the optional *forceComplexType* argument ensuring the
    resulting form is treated as complex.

    Parameters:
    -----------
    arg1 : DFClass or sympy expression
        The differential form or expression to compute the exterior derivative of.
    forceComplexType : bool, optional
        If True, the result is forced to be of complex type, even if the original form is not (default is None).

    Returns:
    --------
    DFClass
        A new differential form representing the exterior derivative of *arg1*.

    Example:
    --------
    >>> from DGCV import createVariables, makeZeroForm, exteriorDerivative
    >>> createVariables('z', 'x', 'y', withVF=True)
    >>> f_Form = makeZeroForm(z,DGCVType='complex')
    >>> exteriorDerivative(f_Form)
    I*d_y+d_x
    """
    # Check that arg1 is a DFClass or convert it into a zero-form
    if not isinstance(arg1, DFClass) and isinstance(arg1, (int, float, sp.Expr)):
        arg1 = makeZeroForm(arg1, DGCVType="complex" if forceComplexType else None)
    elif arg1.DGCVType == "complex":
        forceComplexType = True

    if isinstance(arg1, DFClass):
        typeSet = arg1._varSpace_type
    else:
        raise TypeError(
            "`exteriorDerivative` can only operator of DFClass or scalars like sp.Expr"
        )

    # Helper function to compute the exterior derivative of a zero-form
    def extDerOfZeroForm(arg1, typeSet=None):
        if arg1.DGCVType == "complex" or forceComplexType:
            if typeSet == "real":
                sparseDataLoc = {
                    (j,): sp.diff(allToReal(arg1.coeffsInKFormBasis[0]), arg1.varSpace[j])
                    for j in range(len(arg1.varSpace))
                }
            else:
                sparseDataLoc = {
                    (j,): sp.diff(allToSym(arg1.coeffsInKFormBasis[0]), arg1.varSpace[j])
                    for j in range(len(arg1.varSpace))
                }
        else:
            sparseDataLoc = {
                (j,): sp.diff(arg1.coeffsInKFormBasis[0], arg1.varSpace[j])
                for j in range(len(arg1.varSpace))
            }

        return DFClass(
            arg1.varSpace,
            sparseDataLoc,
            1,
            DGCVType="complex" if forceComplexType else arg1.DGCVType,
        )

    # Handle zero-forms
    if arg1.degree == 0:
        return extDerOfZeroForm(arg1, typeSet=typeSet)

    # Handle higher-degree forms
    minDataLoc = arg1.DFClassDataMinimal
    coeffsTo1Forms = [
        extDerOfZeroForm(makeZeroForm(j[1], varSpace=arg1.varSpace), typeSet=typeSet)
        for j in minDataLoc
    ]
    # Construct the corresponding basis k-forms
    basisOfCoeffs = [
        DFClass(
            arg1.varSpace,
            {tuple(j[0]): 1},
            arg1.degree,
            DGCVType="complex" if forceComplexType else arg1.DGCVType,
        )
        for j in minDataLoc
    ]

    # Multiply the one-forms by the basis and sum them
    return addDF(
        *[coeffsTo1Forms[j] * basisOfCoeffs[j] for j in range(len(minDataLoc))]
    )


def LieDerivativeOf1Form(vf, oneForm):
    """
    Computes the Lie derivative of a 1-form (i.e., DFClass instance) *oneForm* along a vector field (i.e., VFClass instance) *vf*.

    The Lie derivative is computed using Cartan's formula:
        L_v(ω) = i_v(dω) + d(i_v(ω))
    where L_v is the Lie derivative along the vector field *vf*, ω is the 1-form, d is the exterior derivative, and i_v is the interior product with respect to the vector field.

    Parameters:
    -----------
    vf : VFClass
        The vector field along which the Lie derivative is computed.
    oneForm : DFClass
        The 1-form to compute the Lie derivative of.

    Returns:
    --------
    DFClass
        The Lie derivative of the 1-form along the vector field.

    """
    # Compute the Lie derivative using Cartan's formula
    return interiorProductOf2Form(vf, exteriorDerivative(oneForm)) + exteriorDerivative(
        interiorProductOf1Form(vf, oneForm)
    )


def interiorProductOf1Form(vf, oneForm):
    """
    Computes the interior product of a vector field (i.e., VFClass instance) *vf* and a 1-form (i.e., DFClass instance) *oneForm*.

    The interior product (or contraction) of a vector field and a 1-form produces a 0-form, which can be viewed as the result of applying the 1-form to the vector field.

    Parameters:
    -----------
    vf : VFClass
        The vector field to contract with the 1-form.
    oneForm : DFClass
        The 1-form to contract with the vector field.

    Returns:
    --------
    DFClass
        A 0-form representing the result of the interior product of *vf* and *oneForm*.
    """
    varSpaceLoc = tuple(dict.fromkeys(vf.varSpace + oneForm.varSpace))
    return makeZeroForm(oneForm(vf), varSpace=varSpaceLoc, DGCVType=oneForm.DGCVType)


def interiorProductOf2Form(vf, twoForm):
    """
    Computes the interior product of a vector field (i.e., VFClass instance) *vf* and a 2-form (i.e., DFClass instance) *twoForm*.

    The result is a 1-form, expressed as a linear combination of 1-forms in the same variable space.

    Parameters:
    -----------
    vf : VFClass
        The vector field to contract with the 2-form.
    twoForm : DFClass
        The 2-form to contract with the vector field.

    Returns:
    --------
    DFClass
        A 1-form representing the result of the interior product of *vf* and *twoForm*.
    """
    varSpaceLoc = tuple(dict.fromkeys(vf.varSpace + twoForm.varSpace))

    # Access the global symbols directly using _cached_caller_globals
    coeffsLoc = [
        twoForm(vf, _cached_caller_globals["D_" + str(k)]) for k in varSpaceLoc
    ]
    oneFormBasisLoc = [_cached_caller_globals["d_" + str(j)] for j in varSpaceLoc]

    # Return the result as a sum of the contracted 1-form
    return addDF(*[coeffsLoc[j] * oneFormBasisLoc[j] for j in range(len(varSpaceLoc))])


def interiorProduct(vf, kForm):
    r"""
    Computes the interior product of a vector field *vf* with a differential k-form *kForm* where k > 0.

    The interior product, or contraction, reduces the degree of a differential form by 1 and represents
    the action of the vector field on the form. That is, the vector field is "taking one for the team"!

    Parameters:
    -----------
    vf : VFClass
        The vector field to contract with the differential form.
    kForm : DFClass
        A differential k-form (degree > 0) to contract with the vector field.

    Returns:
    --------
    DFClass
        A differential form (degree k-1) representing the result of the interior product.

    Raises:
    -------
    Exception
        If *vf* is not a VFClass instance or *kForm* is not a DFClass instance with degree > 0.

    Example:
    --------
    >>> from DGCV import createVariables, makeZeroForm, interiorProduct
    >>> from sympy import latex
    >>> createVariables('z', 'x', 'y')
    >>> createVariables('v')
    >>> vf = v*D_x + D_y
    >>> omega = d_x * d_y * d_v
    >>> print(latex(interiorProduct(vf, omega)))
    $vd y \wedge d v -d x \wedge d v$
    """
    if [vf.__class__.__name__, kForm.__class__.__name__] != ["VFClass", "DFClass"]:
        raise Exception(
            "Arguments *vf* and *kForm* must have respective types *VFClass* and *DFClass*."
        )

    if kForm.degree == 0:
        raise Exception("Differential form must have degree > 0")

    # Initialize a sparse dictionary to store non-zero entries
    sparse_array_loc = {}

    # Generate the list of indices for the contraction
    indexListLoc = [list(j) for j in chooseOp(range(len(kForm.varSpace)), kForm.degree)]
    indexListFilteredLoc = [
        list(j)
        for j in chooseOp(
            range(len(kForm.varSpace)), kForm.degree, withoutReplacement=True
        )
    ]

    # Access the vector field components from the caller's globals
    VFListLoc = ["D_" + str(j) for j in kForm.varSpace]

    # Compute the interior product for each component, storing non-zero entries in sparse_array_loc
    for j in indexListLoc:
        result = kForm(*([vf] + [_cached_caller_globals[VFListLoc[k]] for k in j[1:]]))
        if j in indexListFilteredLoc and result != 0:
            sparse_array_loc[tuple(j[1:])] = result

    # Return a differential form using the sparse dictionary
    return DFClass(
        kForm.varSpace, sparse_array_loc, kForm.degree - 1, DGCVType=kForm.DGCVType
    )


def LieDerivative(vf, arg):
    r"""
    The Lie derivative, because sometimes differential forms need a little nudge in the right direction.

    Computes the Lie derivative with respect to the vector field (i.e., VFClass instance) *vf* of either another vector field
    or a differential form (i.e., DFClass instance) *arg*.

    The Lie derivative of a vector field is given by the Lie bracket, and for a differential form, it is computed using Cartan's formula:
        L_v(ω) = d(i_v(ω)) + i_v(dω)

    Parameters:
    -----------
    vf : VFClass
        The vector field with respect to which the Lie derivative is computed.
    arg : VFClass or DFClass
        The vector field or differential form to compute the Lie derivative of.

    Returns:
    --------
    VFClass or DFClass
        The Lie derivative of *arg* with respect to *vf*.

    Raises:
    -------
    Exception
        If *vf* is not of type VFClass or *arg* is not of type VFClass/DFClass.

    Example:
    --------
    >>> from DGCV import createVariables, LieDerivative
    >>> from sympy import latex
    >>> createVariables('z', 'x', 'y')
    >>> createVariables('v', withVF=True)
    >>> vf = x * D_x
    >>> omega = d_x * d_y * d_v
    >>> print(latex(LieDerivative(vf, omega)))
    $d x \wedge d y \wedge d v$
    """
    if isinstance(vf, VFClass) and isinstance(arg, (int, float, sp.Expr)):
        return vf(arg)
    # Check for valid argument types
    if [vf.__class__.__name__, arg.__class__.__name__] not in [
        ["VFClass", "DFClass"],
        ["VFClass", "VFClass"],
    ]:
        raise Exception(
            "Arguments *vf* and *arg* must have respective types *VFClass* and *VFClass/DFClass*."
        )

    # Case: Lie derivative of a differential form
    if isinstance(arg, DFClass):
        if arg.degree > 0:
            coeffList = arg.coeffsInKFormBasis
            oneFormList = [
                [_cached_caller_globals[k] for k in j] for j in arg.kFormBasisGenerators
            ]

            # Helper function to operate on each factor of a k-form
            def operateOnKFactor(formList, k):
                firstFactor = sp.prod(formList[0:k])
                secondFactor = LieDerivativeOf1Form(vf, formList[k])
                thirdFactor = sp.prod(formList[k + 1 :])
                return firstFactor * secondFactor * thirdFactor

            # Compute the Lie derivative
            oneFormDerivatives = addDF(
                *[
                    coeffList[j] * operateOnKFactor(oneFormList[j], k)
                    for k in range(arg.degree)
                    for j in range(len(coeffList))
                ]
            )
            coeffDerivatives = addDF(
                *[
                    vf(coeffList[j]) * sp.prod(oneFormList[j])
                    for j in range(len(coeffList))
                ]
            )
            return coeffDerivatives + oneFormDerivatives

        # Case: Lie derivative of a 0-form
        elif arg.degree == 0:
            return DFClass(arg.varSpace, {0: vf(arg)}, 0)

    # Case: Lie derivative of a vector field (Lie bracket)
    return VF_bracket(vf, arg)


############## vector field and differential forms operations with linsolve computations


def decompose(
    obj,
    basis,
    return_parameters=False,
    new_parameters_label=None,
    _pass_error_report=None,
    _hand_off=None,
):
    """
    Decomposes a vector field or differential form as a linear combination of a given `basis` list.

    This function attempts to express the input `obj` (a VFClass or DFClass object) as a linear combination
    of the elements in the provided `basis` list. The `basis` list does not need to be linearly independent,
    and if the decomposition is not unique, the function will parameterize the solution space. Any parameters
    needed are automatically initialized and registered in DGCV's variable management framework (VMF).

    The function carefully handles variable types based on the `DGCVType` attribute of the objects. For objects
    with `DGCVType='complex'`, it dynamically selects whether to perform real coordinate computations or complex
    coordinate computations, depending on the input data. If a canonical variable formatting decision cannot be
    made naturally from the input, the function will return warnings with explanations.

    Parameters
    ----------
    obj : VFClass or DFClass
        The vector field or differential form to decompose.
    basis : list of VFClass or DFClass
        A list of objects (vector fields or differential forms) to decompose `obj` with respect to.
        The class of objects in the `basis` list must match the class of `obj`.
    return_parameters : bool, optional
        If True, the function will return the parameterized solution when the decomposition is not unique
        (default is False). Parameters are initialized with labels registered within the VMF.
    new_parameters_label : str or None, optional
        If `return_parameters` is True and the decomposition is not unique, this label will be used
        to name the new parameter variables. If None, obscure labels will be generated automatically (default is None).
    _pass_error_report : optional
        Internal use parameter for handling error reports in certain edge cases (default is None).

    Returns
    -------
    list
        The coefficients of the linear combination that expresses `obj` in terms of the `basis` list.
        If the decomposition is parameterized, the returned list contains the parameterized solution.

    Raises
    ------
    TypeError
        If the class of `obj` does not match the class of elements in the `basis` list (i.e., both must
        be either VFClass or DFClass), or if objects in the `basis` list have inconsistent `DGCVType` attributes.

    Warnings
    --------
    - A warning is issued if `obj` is not in the span of the provided `basis` list.
    - If the `basis` list is not linearly independent, the decomposition is not unique, and a parameterized
    solution will be returned. The initialized parameters are registered as 'standard variables' in the VMF.

    Remarks
    -------
    - The function dynamically handles objects based on their `DGCVType` attribute. For `DGCVType='complex'`,
    it distinguishes between real and complex coordinate computations, converting the input as needed
    to ensure consistency in formatting. If this decision cannot be determined from the input data,
    the function issues warnings explaining the necessary canonical formatting.

    Example
    -------
    # Decompose a vector field 'vf' as a linear combination of two basis vector fields 'vf1' and 'vf2'
    coeffs, basis_used = decompose(vf, [vf1, vf2])

    # Decompose a differential form 'df' with a parameterized solution due to non-uniqueness
    coeffs, basis_used = decompose(df, [df1, df2, df3], return_parameters=True, new_parameters_label='p')
    """
    if not isinstance(obj, (VFClass, DFClass)):
        raise TypeError(
            "`decompose` only operates on `DFClass` and `VFClass` objects at this time."
        )
    condition1 = isinstance(obj, VFClass) and all(
        [isinstance(j, VFClass) for j in basis]
    )
    condition2 = isinstance(obj, DFClass) and all(
        [isinstance(j, DFClass) for j in basis]
    )
    if condition2:
        basis = [
            j for j in basis if j.degree == obj.degree
        ]  # discard unneeded DFCass elements
    if not condition1 and not condition2:
        raise TypeError(
            "`decompose` needs all objects to be of the same class (i.e., all vector fields VFClass or add diff. forms DFClass)"
        )

    if len(set([j.DGCVType for j in basis])) > 1:
        raise TypeError(
            "`decompose` was given vector fields or differential forms with different `DGCVType` attribute values. Reformat them to all have `DGCVType` 'standard' or 'complex'. The descision whether or not to track complex variable systems in this computation affects the outcome, so `decompose` will not automate attribute homogenization here by design."
        )

    # make _varSpace_type uniform and strip unnecessary variables from varSpace attributes
    if obj.DGCVType == "complex":
        if obj._varSpace_type == "real":
            obj = compressDGCVClass(allToReal(obj))
            basis = [compressDGCVClass(allToReal(j)) for j in basis]
        elif obj._varSpace_type == "complex":
            obj = compressDGCVClass(allToSym(obj))
            basis = [compressDGCVClass(allToSym(j)) for j in basis]
    else:
        obj = compressDGCVClass(obj)
        basis = [compressDGCVClass(j) for j in basis]

    dimLoc = len(basis)
    if dimLoc == 0:
        return [tuple(), tuple()]
    tempLabel = "deco" + create_key()
    variableProcedure(tempLabel, dimLoc, initialIndex=0, _tempVar=retrieve_passkey())
    tempVars = _cached_caller_globals[tempLabel]
    if isinstance(obj, VFClass):
        solObj = compressDGCVClass(
            addVF(
                *[
                    _cached_caller_globals[tempLabel + str(j)] * basis[j]
                    for j in range(dimLoc)
                ]
            )
        )
        if not all([j in solObj.varSpace for j in obj.varSpace]):
            warnings.warn(f"The vector field {obj} is not in the span of {basis}")
            return []
        eqns = (compressDGCVClass(obj - solObj)).coeffs
    if isinstance(obj, DFClass):
        solObj = compressDGCVClass(
            addDF(
                *[
                    _cached_caller_globals[tempLabel + str(j)] * basis[j]
                    for j in range(dimLoc)
                ]
            )
        )
        if not all([j in solObj.varSpace for j in obj.varSpace]):
            warnings.warn(f"The differential form {obj} is not in the span of {basis}. Variable space mis-alignment: {solObj.varSpace} and {obj.varSpace}")
            obj = changeDFBasis(obj, solObj.varSpace)
            return []
        eqns = [j[1] for j in (compressDGCVClass(obj - solObj)).DFClassDataMinimal]

    sol = sp.linsolve(eqns, tempVars)
    if len(list(sol)) == 0:
        printStr = f"`decompose` rolled back its algorithm because `linsolve` failed to solve the equations {eqns} w.r.t. the variables {tempVars}"
        clearVar(*listVar(temporary_only=True), report=False)
        if _pass_error_report == retrieve_passkey():
            return printStr
        else:
            raise Exception(printStr)
    solDict = dict(zip(tempVars, list(sol)[0]))
    freeVar = [t for t in tempVars if t == solDict[t]]
    if return_parameters and len(freeVar) > 0:
        if isinstance(new_parameters_label, str):
            newLabel = new_parameters_label
            variableProcedure(newLabel, len(freeVar))
        else:
            newLabel = "NULL" + create_key()
            variableProcedure(newLabel, len(freeVar), _obscure=None)
            warnings.warn(
                f"The provided object list to decompose w.r.t. is not linearly independent, so variables with intintionally obscure labels where created to parameteraze the solution space. To have the new variable's be assigned a particular label instead, use `new_parameters_label=True`.\n The new labels are: {list(_cached_caller_globals[newLabel])} \n Use that list to substitute the obscure labels for nicer ones as needed."
            )
        subDict = dict(zip(freeVar, _cached_caller_globals[newLabel]))
        return_list = [sp.sympify(j).subs(subDict) for j in list(sol)[0]]
        clearVar(*listVar(temporary_only=True), report=False)
        return return_list, basis
    else:
        if len(freeVar) > 0 and _hand_off != retrieve_passkey():
            warnings.warn(
                "The provided object list to decompose w.r.t. is not linearly independent, so the decomposition might not be unique. `decompostion` will still return a solution. To get all solutions with parameters set `return_parameters=True`."
            )
        subsList = [(t, 0) for t in freeVar]
        clearVar(*listVar(temporary_only=True), report=False)
        return [sp.sympify(j).subs(subsList) for j in list(sol)[0]], basis


def get_coframe(VFList):
    """
    Constructs a coframe (set of differential forms) corresponding to a given list of vector fields.

    This function computes a coframe, a set of differential forms, for the input list of vector fields `VFList`.
    Each differential form in the coframe evaluates to 1 when applied to its corresponding vector field and
    to 0 when applied to all other vector fields in the list. The function dynamically handles both real and
    holomorphic coordinate systems for vector fields.

    For vector fields with `DGCVType='complex'`, `get_coframe` determines whether evaluations should default
    to real or holomorphic coordinates based on the first vector field in the list, and all other vector fields
    are converted accordingly. If such conversions occur, `get_coframe` issues warnings
    explaining its formatting choice.

    Parameters
    ----------
    VFList : list of VFClass
        A list of vector fields for which the coframe will be constructed. All vector fields in the list must
        have the same `DGCVType` attribute (i.e., 'standard' or 'complex'), and for the `DGCVType='complex'`
        case coordinate systems used in their definitions should be consistent (real or holomorphic).

    Returns
    -------
    list of DFClass
        The coframe, represented as a list of differential forms corresponding to the input vector fields.
        Each differential form evaluates to 1 on the corresponding vector field and 0 on the others.

    Raises
    ------
    TypeError
        If the vector fields in `VFList` have inconsistent `DGCVType` attributes or if the list contains
        linearly dependent vector fields that prevent the construction of a coframe.

    Warnings
    --------
    - If the vector fields do not span the minimal coordinate space containing variables in their definitions,
    the function returns a non-unique "coframe" and issues
    a warning.
    - If some vector fields are defined with respect to real coordinates and others with respect to holomorphic
    coordinates, the function automatically converts them all to a consistent coordinate system based on the
    first vector field in the list, and a warning is issued.

    Example
    -------
    >>> from DGCV import createVariables, get_coframe
    >>> createVariables('z', 'x', 'y', 3, default_var_format='real')
    >>> vfList = [D_z1 + z1**3 * D_BARz1, D_BARz1 + BARz1**3 * D_z1]
    >>> coframe = get_coframe(vfList)
    >>> print(coframe)

    [((x1**3 - 3*I*x1**2*y1 - 3*x1*y1**2 + I*y1**3 - 1)/(x1**6 + 3*x1**4*y1**2 + 3*x1**2*y1**4 + y1**6 - 1))*d_x1 +
    ((-I*x1**3 - 3*x1**2*y1 + 3*I*x1*y1**2 + y1**3 - I)/(x1**6 + 3*x1**4*y1**2 + 3*x1**2*y1**4 + y1**6 - 1))*d_y1,
    ((x1**3 + 3*I*x1**2*y1 - 3*x1*y1**2 - I*y1**3 - 1)/(x1**6 + 3*x1**4*y1**2 + 3*x1**2*y1**4 + y1**6 - 1))*d_x1 +
    ((I*x1**3 - 3*x1**2*y1 - 3*I*x1*y1**2 + y1**3 + I)/(x1**6 + 3*x1**4*y1**2 + 3*x1**2*y1**4 + y1**6 - 1))*d_y1]
    """
    dimLoc = len(VFList)
    vr = get_variable_registry()
    if len(set([j.DGCVType for j in VFList])) > 1:
        raise TypeError(
            "`coframe` was given a set of vector fields with different `DGCVType` attribute values. Reformat the vector fields to all have `DGCVType` 'standard' or 'complex'"
        )
    elif VFList[0]._varSpace_type == "real":
        varFormat = "real"
        VFList = [allToReal(j) for j in VFList]
        if any([j._varSpace_type == "complex" for j in VFList[1:]]):
            warnings.warn(
                "`coframe` was given a list of `DGCVType='complex'` vector fields in which some are defined w.r.t. to real coordinates and others w.r.t. holomorphic coordinates. To find a common coordinate system `coframe` converted everything to the coordinate type of the first provided VF, which is `real`. Tip: if holomorphic coordinates were prefered then convert the provided VF to holom. format before giving them to `coframe` (it is even enough to do this for just the first VF in the list). Use functions like `allToHol` and `allToSym` to achieve this."
            )
    elif VFList[0]._varSpace_type == "complex":
        varFormat = "complex"
        VFList = [allToSym(j) for j in VFList]
        if any([j._varSpace_type == "real" for j in VFList[1:]]):
            warnings.warn(
                "`coframe` was given a list of `DGCVType='complex'` vector fields in which some are defined w.r.t. to real coordinates and others w.r.t. holomorphic coordinates. To find a common coordinate system `coframe` converted everything to the coordinate type of the first provided VF, which is `holomorphic`. Tip: if real coordinates were prefered then convert the provided VF to real format before giving them to `coframe` (it is even enough to do this for just the first VF in the list). The function `allToReal` is one way to do such conversion."
            )
    else:
        varFormat = "standard"
    varSpaceLoc = tuple(
        set.union(*[set(d.keys()) for d in [minimalVFDataDict(j) for j in VFList]])
    )
    if len(varSpaceLoc) > dimLoc:
        warnings.warn(
            'Fewer vector fields were given than the minimal number of variables needed to define them in the given coordinate representations, so the coframe solution returned will not be unique. A "coframe" will be returned anyway. Add a complimentary set of VF to specifify a unique solution. Example: if working with holomophic VF or real VF expressions in coordinates from a complex system, toss in the coordinate antiholomorphic VF or coordinate purely imaginary (e.g., D_z-D_BARz) to fill out a basis.\n If this message comes as a surprise, it may be that you are providing real VF of complex type (like, D_x or D_y where x and y are real/im. parts of a complex variable system) but your VF were defined w.r.t. holomorphic coordinates rather than real (so internally, D_x is regarded as D_z+D_BARz etc.). Similarly, holomorphic VF defined w.r.t. real coordinates could result in the same suprise, as internally 2*D_z is then regarded D_x-I*D_y etc. Use functions like `allToReal`, `allToHol`, and `allToSym`, for fine control of these properties.'
        )
    if len(varSpaceLoc) < dimLoc:
        raise TypeError(
            "`coframe` cannot find a coframe because the provided vector fields are not linearly independent"
        )

    def getDF(key, parentsDict):
        keyStr = str(key)
        for var in parentsDict:
            if keyStr in parentsDict[var]["variable_relatives"]:
                if parentsDict[var]["variable_relatives"][keyStr]["DFClass"] is None:
                    raise TypeError(
                        "One of the provided vector fields was defined over coordinates including at least one variable that was initialized without a corresponding coordinate vector field registered in the DGCV variable management framework. Suggestion: use DGCV variable creation functions to initialize variables. Use `createVariables(--,withVF=True)` or `createVariables(--,complex=True)` to initialize all variables that the vector fields are defined w.r.t."
                    )
                else:
                    return parentsDict[var]["variable_relatives"][keyStr]["DFClass"]

    if varFormat == "standard":
        dfBasis = [
            _remove_complex_handling(
                getDF(
                    j, vr["standard_variable_systems"] | vr["complex_variable_systems"]
                )
            )
            for j in varSpaceLoc
        ]
    elif varFormat == "complex":
        dfBasis = [
            allToSym(getDF(j, vr["complex_variable_systems"])) for j in varSpaceLoc
        ]
    elif varFormat == "real":
        dfBasis = [
            allToReal(getDF(j, vr["complex_variable_systems"])) for j in varSpaceLoc
        ]
    tempLabel = "coframe" + create_key()
    tempVars = []
    solDFs = []
    eqns = []
    for j in range(dimLoc):
        variableProcedure(
            tempLabel + str(j) + "_", len(dfBasis), _tempVar=retrieve_passkey()
        )
        varListLoc = list(_cached_caller_globals[tempLabel + str(j) + "_"])
        tempVars = tempVars + varListLoc
        joinList = [varListLoc[k] * dfBasis[k] for k in range(len(dfBasis))]
        newDF = addDF(*joinList)
        solDFs = solDFs + [newDF]
        newEqns = [
            newDF(VFList[k]) - 1 if k == j else newDF(VFList[k]) for k in range(dimLoc)
        ]
        eqns = eqns + newEqns

    sol = sp.linsolve(eqns, tempVars)
    if len(list(sol)) == 0:
        warnings.warn("linsolve failed. Trying solve.")
        sol = sp.solve(eqns, tempVars, dict=True)
        if len(sol) == 0:
            printStr = f"`coframe` rolled back its algorithm because `linsolve` and `solve` failed to solve the equations {eqns} w.r.t. the variables {tempVars}"
            clearVar(*listVar(temporary_only=True), report=False)
            raise Exception(printStr)
        else:
            solDict = sol[0]
            flagSolveLogic = True
    else:
        solDict = dict(zip(tempVars, list(sol)[0]))
        flagSolveLogic = False
    if flagSolveLogic:
        freeVar = []
    else:
        freeVar = [t for t in tempVars if t == solDict[t]]
    if len(freeVar) > 0:
        subsList = [(t, 0) for t in freeVar]
        coframeSol = [j.subs(solDict).subs(subsList) for j in solDFs]
        clearVar(*listVar(temporary_only=True), report=False)
    else:
        coframeSol = [j.subs(solDict) for j in solDFs]
        clearVar(*listVar(temporary_only=True), report=False)

    return coframeSol


def annihilator(
    objList,
    coordinate_Space,
    control_distribution=None,
    _pass_error_report=None,
    allow_div_by_zero=False,
):
    """
    Finds annihilators for a given list of vector fields or differential forms.

    This function computes objects that "annihilate" the provided list of vector fields or differential forms.
    An annihilator is either the span of differential forms that evaluate to zero on each vector field in the list,
    or vector fields whose interior product annihilates each differential form in the list. `annihilator` dynamically
    handles both real and holomorphic coordinate systems and can convert between them as needed. Additionally,
    solutions can be constrained to a given control distribution by using the `control_distribution` keyword.

    Parameters
    ----------
    objList : list of VFClass or DFClass
        A list of vector fields or differential forms for which the annihilator will be computed. All objects
        in the list must be of the same class (either all vector fields or all differential forms) and have
        consistent `DGCVType` attributes (i.e., 'standard' or 'complex').
    coordinate_Space : list, tuple, or set
        A collection of variables that define the coordinate system in which the annihilator is to be computed.
    allow_div_by_zero : bool, optional
        If True, allows the annihilator to be returned without scaling to avoid division by zero (default is False).
        Scaling to avoid division has more computational overhead but typically simplifies output.
    _pass_error_report : optional
        Internal use parameter for handling error reports in certain edge cases (default is None).

    Returns
    -------
    list
        A list of differential forms (if vector fields were provided) or vector fields (if differential forms were
        provided) that annihilate the objects in `objList`.

    Raises
    ------
    TypeError
        If the objects in `objList` are not all of the same type (i.e., all vector fields or all differential forms),
        or if the `coordinate_Space` is not a valid list, tuple, or set.

    Warnings
    --------
    - If the objects in `objList` are defined with inconsistent coordinate system types (real vs. holomorphic), the
    function converts them to a consistent coordinate system and issues a warning.

    Example
    -------
    >>> from DGCV import createVariables, annihilator, exteriorDerivative, complex_struct_op, Del, DelBar, allToReal
    >>> createVariables('z', 'x', 'y', 4, initialIndex=0)
    >>> rho = (x1*x2 + x1**2*x3 - x0)  # A defining equation for a real hypersurface M in C^4
    >>> d_rho = exteriorDerivative(rho)  # Its differential will annihilate TM
    >>> print(d_rho)

    (2*x1*x3 + x2)*d_x1 + x1*d_x2 + x1**2*d_x3 - 1*d_x0

    >>> dfList = [d_rho]
    >>> TMbasis = annihilator(dfList, x+y)  # Use annihilator to compute the tangent bundle TM
    >>> TMbasis

    [(16*x1*x3 + 8*x2)*D_x0 + 8*D_x1, 8*x1*D_x0 + 8*D_x2, 8*x1**2*D_x0 + 8*D_x3, 8*D_y0, 8*D_y1, 8*D_y2, 8*D_y3]

    >>> J_of_TMbasis = [complex_struct_op(vf) for vf in TMbasis]  # Get the image of TM under the complex structure operator.
    >>> J_of_TMbasis

    [(16*x1*x3 + 8*x2)*D_y0 + 8*D_y1, 8*x1*D_y0 + 8*D_y2, 8*x1**2*D_y0 + 8*D_y3, -8*D_x0, -8*D_x1, -8*D_x2, -8*D_x3]

    >>> CR_distribution = annihilator(annihilator(J_of_TMbasis, x+y) + annihilator(TMbasis, x+y), x+y)
    >>> CR_distribution  # Use annihilator to get the CR distribution, which is the intersection of TM with CTM

    [(16*x1*x3 + 8*x2)*D_x0 + 8*D_x1, 8*x1*D_x0 + 8*D_x2, 8*x1**2*D_x0 + 8*D_x3,
    (16*x1*x3 + 8*x2)*D_y0 + 8*D_y1, 8*x1*D_y0 + 8*D_y2, 8*x1**2*D_y0 + 8*D_y3]

    >>> LeviForm = allToReal(Del(DelBar(rho)))  # Apply Dolbeault operators to represent the Levi form
    >>> print(LeviForm._repr_latex_())

    # Output: <LeviForm in LaTeX formatted plain text>

    >>> Levi_kernel = annihilator([LeviForm], x+y, control_distribution=CR_distribution)
    >>> Levi_kernel  # annihilator reveals that the Levi form has a real 2-d. kernel

    [-64*x1**2*D_x0 - 128*x1*D_x2 + 64*D_x3, -64*x1**2*D_y0 - 128*x1*D_y2 + 64*D_y3]

    >>> not_the_Levi_kernel = annihilator([LeviForm], x+y)
    >>> not_the_Levi_kernel  # Without constraining annihilator to the CR distribution, it finds a kernel that is too large.

    [8*D_x0, -16*x1*D_x2 + 8*D_x3, 8*D_y0, -16*x1*D_y2 + 8*D_y3]
    """
    if not isinstance(coordinate_Space, (list, tuple, set)):
        raise TypeError(
            "`annihilator` needs the provided coordinate space (i.e., second argument) to be a list/tuple/set containing distinct variables. Tip: Use `createVariables(--,withVF=True)` or `createVariables(--,complex=True)` to initialize all variables with corresponding coordinate VF and DF."
        )
    if control_distribution is None:
        control_distribution = []
    elif not isinstance(control_distribution, (list, tuple, set)):
        raise TypeError(
            "`annihilator` needs the (optional) control distribution to be a list/tuple/set containing distinct VFClass/DFClass objects, if provided."
        )
    else:
        control_distribution = list(control_distribution)
    condition1 = all([isinstance(j, VFClass) for j in objList]) and all(
        [isinstance(j, DFClass) for j in control_distribution]
    )
    condition2 = all([isinstance(j, DFClass) for j in objList]) and all(
        [isinstance(j, VFClass) for j in control_distribution]
    )
    if condition2:
        i = 0
        while i < len(objList):
            j = objList[i]
            if j.degree == 0:
                if not j.varSpace:
                    raise TypeError(
                        "`annihilator` recieved a 0-form with trivial variable space."
                    )
                elif not j.is_zero():
                    return VFClass(coordinate_Space, [0])
                else:
                    objList.pop(i)
            else:
                i += 1

    if not condition1 and not condition2:
        raise TypeError(
            "`annihilator` only operates on lists/tuples that contain either all vector fields VFClass or all diff. forms DFClass)If providing a control distribution (optional) to find solutions within, its elements' class must compliment that of the objects to annihilate."
        )

    vr = get_variable_registry()

    if condition1:
        if len(set([j.DGCVType for j in objList + control_distribution])) > 1:
            raise TypeError(
                "`annihilator` was given a set of vector fields (and possibly DFClass controls) with different `DGCVType` attribute values. Reformat the vector fields (and controls if provided) to all have `DGCVType` 'standard' or 'complex'"
            )
        elif objList[0]._varSpace_type == "real":
            varFormat = "real"
            varSpaceLoc = _format_complex_coordinates(
                coordinate_Space, default_var_format="real"
            )
            objList = [allToReal(j) for j in objList]
            control_distribution = [allToReal(j) for j in control_distribution]
            if any([j._varSpace_type == "complex" for j in objList[1:]]):
                warnings.warn(
                    "`annihilator` was given a list of `DGCVType='complex'` vector fields or DFClass controls in which some are defined w.r.t. to real coordinates and others w.r.t. holomorphic coordinates. To find a common coordinate system `annihilator` converted everything to the coordinate type of the first provided VF, which is `real`. Tip: if holomorphic coordinates were prefered then convert the provided VF to holom. format before giving them to `annihilator` (it is even enough to do this for just the first VF in the list). Use functions like `allToHol` and `allToSym` to achieve this."
                )
        elif objList[0]._varSpace_type == "complex":
            varFormat = "complex"
            varSpaceLoc = _format_complex_coordinates(
                coordinate_Space, default_var_format="complex"
            )
            objList = [allToSym(j) for j in objList]
            control_distribution = [allToSym(j) for j in control_distribution]
            if any([j._varSpace_type == "real" for j in objList[1:]]):
                warnings.warn(
                    "`annihilator` was given a list of `DGCVType='complex'` vector fields or DFClass controls in which some are defined w.r.t. to real coordinates and others w.r.t. holomorphic coordinates. To find a common coordinate system `annihilator` converted everything to the coordinate type of the first provided VF, which is `holomorphic`. Tip: if real coordinates were prefered then convert the provided VF to real format before giving them to `annihilator` (it is even enough to do this for just the first VF in the list). The function `allToReal` is one way to do such conversion."
                )
        else:
            varFormat = "standard"
            varSpaceLoc = coordinate_Space

        if control_distribution == []:

            def getDF(key, parentsDict):
                keyStr = str(key)
                for var in parentsDict:
                    if keyStr in parentsDict[var]["variable_relatives"]:
                        if (
                            parentsDict[var]["variable_relatives"][keyStr]["DFClass"]
                            is None
                        ):
                            raise TypeError(
                                "One of the provided vector fields was defined over coordinates including at least one variable that was initialized without a corresponding coordinate vector field registered in the DGCV variable management framework. Suggestion: use DGCV variable creation functions to initialize variables. Use `createVariables(--,withVF=True)` or `createVariables(--,complex=True)` to initialize all variables that the vector fields are defined w.r.t."
                            )
                        else:
                            return parentsDict[var]["variable_relatives"][keyStr][
                                "DFClass"
                            ]

            if varFormat == "standard":
                dfBasis = [
                    _remove_complex_handling(
                        getDF(
                            j,
                            vr["standard_variable_systems"]
                            | vr["complex_variable_systems"],
                        )
                    )
                    for j in varSpaceLoc
                ]
            elif varFormat == "complex":
                dfBasis = [
                    allToSym(getDF(j, vr["complex_variable_systems"]))
                    for j in varSpaceLoc
                ]
            elif varFormat == "real":
                dfBasis = [
                    allToReal(getDF(j, vr["complex_variable_systems"]))
                    for j in varSpaceLoc
                ]
        else:
            dfBasis = control_distribution

        tempLabel = "anni" + create_key()

        dimLoc = len(objList)
        variableProcedure(tempLabel, len(dfBasis), _tempVar=retrieve_passkey())
        tempVars = list(_cached_caller_globals[tempLabel])
        solDF = addDF(*[tempVars[k] * dfBasis[k] for k in range(len(dfBasis))])
        eqns = [solDF(objList[k]) for k in range(dimLoc)]

        sol = sp.linsolve(eqns, tempVars)

        if len(list(sol)) == 0:
            printStr = f"`annihilator` rolled back its algorithm because `linsolve` failed to solve the equations {eqns} w.r.t. the variables {tempVars}"
            clearVar(*listVar(temporary_only=True), report=False)
            if _pass_error_report == retrieve_passkey():
                return printStr
            else:
                raise Exception(printStr)
        solDict = dict(zip(tempVars, list(sol)[0]))
        solDF = solDF.subs(solDict)
        freeVar = [t for t in tempVars if t == solDict[t]]
        if len(freeVar) == 0:
            solDFs = [solDF]
        else:
            solDFs = [
                solDF.subs({j: 1}).subs([(t, 0) for t in freeVar]) for j in freeVar
            ]
        clearVar(*listVar(temporary_only=True), report=False)

        if allow_div_by_zero:
            return solDFs
        else:

            def rescaleCD(df):
                data = df.DFClassDataDict
                denoms = []
                for j in data:
                    denoms += [sp.denom(data[j])]
                scale = sp.simplify(sum(denoms))
                return sp.simplify(scale * df)

            return [rescaleCD(df) for df in solDFs]

    if condition2:
        if len(set([j.DGCVType for j in objList + control_distribution])) > 1:
            raise TypeError(
                "`annihilator` was given a set of differential forms (and possibly VFClass controls) with different `DGCVType` attribute values. Reformat them to all have `DGCVType` 'standard' or 'complex'"
            )
        elif objList[0]._varSpace_type == "real":
            varFormat = "real"
            varSpaceLoc = _format_complex_coordinates(
                coordinate_Space, default_var_format="real"
            )
            control_distribution = [allToReal(j) for j in control_distribution]
            if any([j._varSpace_type == "complex" for j in objList[1:]]):
                warnings.warn(
                    "`annihilator` was given a list of `DGCVType='complex'` differential forms or VFClass controls in which some are defined w.r.t. to real coordinates and others w.r.t. holomorphic coordinates. To find a common coordinate system `annihilator` converted everything to the coordinate type of the first provided DF, which is `real`. Tip: if holomorphic coordinates were prefered then convert the provided DF to holom. format before giving them to `annihilator` (it is even enough to do this for just the first DF in the list). Use functions like `allToHol` and `allToSym` to achieve this."
                )
                objList = [allToReal(j) for j in objList]
        elif objList[0]._varSpace_type == "complex":
            varFormat = "complex"
            varSpaceLoc = _format_complex_coordinates(
                coordinate_Space, default_var_format="complex"
            )
            control_distribution = [allToSym(j) for j in control_distribution]
            if any([j._varSpace_type == "real" for j in objList[1:]]):
                warnings.warn(
                    "`annihilator` was given a list of `DGCVType='complex'` differential forms or VFClass controls in which some are defined w.r.t. to real coordinates and others w.r.t. holomorphic coordinates. To find a common coordinate system `annihilator` converted everything to the coordinate type of the first provided DF, which is `holomorphic`. Tip: if real coordinates were prefered then convert the provided DF to real format before giving them to `annihilator` (it is even enough to do this for just the first DF in the list). The function `allToReal` is one way to do such conversion."
                )
                objList = [allToSym(j) for j in objList]
        else:
            varFormat = "standard"
            varSpaceLoc = coordinate_Space
        objList = [changeDFBasis(j, varSpaceLoc) for j in objList]

        if control_distribution == []:

            def getVF(key, parentsDict):
                keyStr = str(key)
                for var in parentsDict:
                    if keyStr in parentsDict[var]["variable_relatives"]:
                        if (
                            parentsDict[var]["variable_relatives"][keyStr]["VFClass"]
                            is None
                        ):
                            raise TypeError(
                                "One of the provided differential forms was defined over coordinates including at least one variable that was initialized without a corresponding coordinate 1-form registered in the DGCV variable management framework. Suggestion: use DGCV variable creation functions to initialize variables. Use `createVariables(--,withVF=True)` or `createVariables(--,complex=True)` to initialize all variables that the differential forms are defined w.r.t."
                            )
                        else:
                            return parentsDict[var]["variable_relatives"][keyStr][
                                "VFClass"
                            ]

            if varFormat == "standard":
                vfBasis = [
                    _remove_complex_handling(
                        getVF(
                            j,
                            vr["standard_variable_systems"]
                            | vr["complex_variable_systems"],
                        )
                    )
                    for j in varSpaceLoc
                ]
            elif varFormat == "complex":
                vfBasis = [
                    allToSym(getVF(j, vr["complex_variable_systems"]))
                    for j in varSpaceLoc
                ]
            elif varFormat == "real":
                vfBasis = [
                    allToReal(getVF(j, vr["complex_variable_systems"]))
                    for j in varSpaceLoc
                ]
        else:
            vfBasis = control_distribution

        tempLabel = "anni" + create_key()

        dimLoc = len(objList)
        variableProcedure(tempLabel, len(vfBasis), _tempVar=retrieve_passkey())
        tempVars = list(_cached_caller_globals[tempLabel])
        solVF = addVF(*[tempVars[k] * vfBasis[k] for k in range(len(vfBasis))])
        eqns = sum(
            [
                [L[1] for L in interiorProduct(solVF, objList[k]).DFClassDataMinimal]
                for k in range(dimLoc)
            ],
            [],
        )

        sol = sp.linsolve(eqns, tempVars)

        if len(list(sol)) == 0:
            printStr = f"`annihilator` rolled back its algorithm because `linsolve` failed to solve the equations {eqns} w.r.t. the variables {tempVars}"
            clearVar(*listVar(temporary_only=True), report=False)
            if _pass_error_report == retrieve_passkey():
                return printStr
            else:
                raise Exception(printStr)
        solDict = dict(zip(tempVars, list(sol)[0]))
        solVF = solVF.subs(solDict)
        freeVar = [t for t in tempVars if t == solDict[t]]
        if len(freeVar) == 0:
            solVFs = [solVF]
        else:
            solVFs = [
                solVF.subs({j: 1}).subs([(t, 0) for t in freeVar]) for j in freeVar
            ]
        clearVar(*listVar(temporary_only=True), report=False)
        if allow_div_by_zero:
            return solVFs
        else:

            def rescaleCD(vf):
                data = vf.coeffs
                denoms = [sp.denom(j) for j in data]
                scale = sp.simplify(sum(denoms))
                return sp.simplify(scale * vf)

            return [rescaleCD(vf) for vf in solVFs]


# class structureEquations(Basic):

#     def __new__(cls, basis, structureFunctionsDict, connectionFormsDict, torsion = None):
#         # Call Basic.__new__ with only the positional arguments
#         obj = Basic.__new__(cls, basis, structureFunctionsDict)
#         return obj

#     def __init__(self, basis, structureFunctionsDict, connectionFormsDict, torsion = None):
#         self.basis = basis
#         self.structureFunctionsDict = structureFunctionsDict
#         self.connectionFormsDict = connectionFormsDict
#         self.torsion = torsion

#     def __str__(self):
#         result = "Structure Functions:\n"
#         for basis_elem, inner_dict in self.structureFunctionsDict.items():
#             result += f"{basis_elem} : {inner_dict}\n"
#         return result

#     def _repr_latex_(self):
#         latex_result = r"\begin{align*}"
#         dummy_labels = [f"\\_e_{{{i+1}}}" for i in range(len(self.basis))]  # Create dummy labels for the 1-forms

#         # Loop through the structure functions dict
#         for i, (basis_elem, two_form_dict) in enumerate(self.structureFunctionsDict.items()):
#             latex_result += f"d ({dummy_labels[i]}) &= "  # Start the equation with the 1-form dummy label

#             # Build the 2-form sum
#             two_form_sum = []
#             for (j, k), coeff in two_form_dict.items():
#                 two_form_label = f"{dummy_labels[j]} \\wedge {dummy_labels[k]}"
#                 term = f"{LaTeX(coeff)} \\cdot {two_form_label}"
#                 two_form_sum.append(term)

#             # Join the terms in the sum
#             latex_result += " + ".join(two_form_sum) + r" \\"

#         latex_result += r"\end{align*}"
#         return latex_result

#     def _eval_simplify(self, **kwargs):
#         # Loop through the structure functions and apply simplify to each coefficient
#         simplified_structureFunctions = {}

#         for basis_elem, two_form_dict in self.structureFunctionsDict.items():
#             simplified_inner_dict = {}
#             for key, val in two_form_dict.items():
#                 # Apply SymPy's simplify to the value
#                 simplified_val = simplify(val, **kwargs)
#                 simplified_inner_dict[key] = simplified_val
#             simplified_structureFunctions[basis_elem] = simplified_inner_dict

#         # Return a new instance of structureEquations with the simplified data
#         return structureEquations(self.basis, simplified_structureFunctions)

#     def subs(self, *args, **kwargs):
#         # Create a new dictionary to hold the substituted structure functions
#         substituted_structureFunctions = {}

#         # Iterate over the structure functions dictionary
#         for basis_elem, two_form_dict in self.structureFunctionsDict.items():
#             # Initialize an inner dictionary for the new substituted values
#             substituted_inner_dict = {}

#             # Iterate over the inner dictionary (2-form indices and their coefficients)
#             for key, val in two_form_dict.items():
#                 # Apply sympy's substitution method
#                 substituted_val = sp.sympify(val).subs(*args, **kwargs)
#                 substituted_inner_dict[key] = substituted_val

#             # Store the substituted inner dictionary in the outer dictionary
#             substituted_structureFunctions[basis_elem] = substituted_inner_dict

#         # Return a new instance of the structureEquations class with the substituted data
#         return structureEquations(self.basis, substituted_structureFunctions)


# def get_structureEquations(dfList, torsion = None):
#     if not isinstance(dfList,(list,tuple)):
#         raise TypeError('`structureEquations needs the coframe to be a list/tuple of DFClass objects.')
#     elif not all([isinstance(j,DFClass) for j in dfList]):
#         raise TypeError('`structureEquations needs the coframe to be a list/tuple of DFClass objects.')
#     elif not all([j.degree==1 for j in dfList]):
#         raise TypeError('`structureEquations was given a list/tuple k-forms for the coframe whose degrees are not all one.')
#     if torsion == None:
#         df = get_DF(dfList[0].varSpace[0])[0]
#         Torsion = [0*df*df]*len(dfList)
#     elif not isinstance(torsion,(list,tuple)):
#         raise TypeError('If providing torsion `structureEquations` needs it to be a list of 2-forms.')
#     elif not all([isinstance(j,DFClass) for j in Torsion]):
#         raise TypeError('If providing torsion `structureEquations` needs it to be a list of 2-forms.')
#     elif not all([j.degree==2 for j in Torsion]):
#         raise TypeError('If providing torsion `structureEquations` needs it to be a list of 2-forms.')
#     dfList = list(dfList)
#     Torsion = list(Torsion)

#     if len(set([j.DGCVType for j in dfList]))>1:
#         raise TypeError('`get_structureEquations` was given differential forms with different `DGCVType` attribute values. Reformat them to all have `DGCVType` \'standard\' or \'complex\'. The descision whether or not to track complex variable systems in this computation affects the outcome, so `get_structureEquations` will not automate attribute homogenization here by design.')

#     if Torsion != None:
#         if len(set([j.DGCVType for j in dfList+Torsion]))>1:
#             raise TypeError('`get_structureEquations` was given differential forms with different `DGCVType` attribute values. Reformat them to all have `DGCVType` \'standard\' or \'complex\'. The descision whether or not to track complex variable systems in this computation affects the outcome, so `get_structureEquations` will not automate attribute homogenization here by design.')

#     # make _varSpace_type uniform
#     if dfList[0].DGCVType=='complex':
#         if dfList[0]._varSpace_type=='real':
#             Torsion = [allToReal(j) for j in Torsion]
#             if any([j._varSpace_type!='real' for j in dfList]):
#                 dfList = [allToReal(j) for j in dfList]
#         elif dfList[0]._varSpace_type=='complex':
#             Torsion = [allToSym(j) for j in Torsion]
#             if any([j._varSpace_type!='complex' for j in dfList]):
#                 dfList = [allToSym(j) for j in dfList]

#     unraveler = dict()
#     twoForms = []
#     count = 0
#     if len(dfList)>1:
#         indexList=[(j,k) for j in range(len(dfList)) for k in range(len(dfList))]
#     else:
#         indexList=[(0,0)]
#     for index in indexList:
#         unraveler[count] = index
#         count = count+1
#         twoForms = twoForms +[dfList[index[0]]*dfList[index[1]]]

#     structureFunctions=dict()
#     for j in range(len(dfList)):
#         df=dfList[j]
#         extdf = Torsion[j]-exteriorDerivative(df)
#         fetchSol = decompose(extdf,twoForms,_hand_off = retrieve_passkey())
#         if isinstance(fetchSol,str):
#             raise ValueError(f'`structureEquations` failed. It appears the exterior derivative of {df} is not in the span of the provided 1-forms\' wedge products. The related error report from DGCV `decompose` is as follows:\n {fetchSol}')
#         else:
#             structureFunctions[j] = {unraveler[k]:fetchSol[0][k] for k in range(len(fetchSol[0]))}

#     connectionForms=dict()
#     for j in structureFunctions:
#         for k in range(len(dfList)):
#             omega = addDF([structureFunctions[j][l]*dfList[l[0]] for l in structureFunctions[j] if l[1]==k])
#             connectionForms[(j,k)] = omega


#     if torsion==None:
#         tval=None
#     else:
#         tval = Torsion

#     return structureEquations(dfList,structureFunctions, connectionForms, torsion = Torsion)



"""
DGCV: Differential Geometry with Complex Variables

This module provides tools specific to CR (Cauchy-Riemann) geometry within the DGCV package. 
It includes functions for constructing CR hypersurfaces and  computing symmetries.

Key Functions:
    - tangencyObstruction(): Computes the tangency obstruction for a holomorphic vector field's 
      real part to be tangent to a CR hypersurface.
    - weightedHomogeneousVF(): Creates a general weighted homogeneous vector field in a 
      specified coordinate space.
    - findWeightedCRSymmetries(): Attempts to find infinitesimal symmetries of a weighted CR 
      hypersurface.
    - model2Nondegenerate(): Builds the defining equation for a 2-nondegenerate model 
      hypersurface.

Author: David Sykes (https://github.com/YikesItsSykes)

Dependencies:
    - sympy

License:
    MIT License
"""

############## dependencies
############## CR geometry
from functools import reduce

import sympy as sp

from ._safeguards import create_key, retrieve_passkey
from .config import _cached_caller_globals, get_variable_registry
from .dgcv_core import (
    VFClass,
    addVF,
    allToReal,
    clearVar,
    holToReal,
    holVF_coeffs,
    listVar,
    realPartOfVF,
    scaleVF,
    symToReal,
    variableProcedure,
)
from .polynomials import createPolynomial
from .solvers import solve_DGCV
from .vector_fields_and_differential_forms import assembleFromHolVFC


def tangencyObstruction(vf, CR_defining_expr, graph_variable, simplify=False, *args):
    """
    Computes the tangency obstruction for a holomorphic vector field with respect to a CR hypersurface.

    This function computes the Lie derivative of the real part of a holomorphic vector field applied
    to the defining equation of a CR hypersurface and then substitutes the defining equation into the result.
    The output is zero (up to simplification) if and only if the vector field is a symmetry of the hypersurface.

    Parameters:
    -----------
    vf : VFClass
        A holomorphic vector field in the complex coordinate 
    CR_defining_expr : sympy expression
        A defining function of the CR hypersurface, expressed in holomorphic or real variables.
    graph_variable : sympy symbol
        The real variable whose value is set equal to the defining function to define the hypersurface.
    simplify : bool, optional
        If True, applies sp.simplify to the final result.
    *args:
        Additional arguments (unused).

    Returns:
    --------
    sympy expression
        The tangency obstruction.

    Raises:
    -------
    TypeError
        If the first argument is not a VFClass instance with DGCVType='complex'.
    """
    if not (isinstance(vf, VFClass) and vf.DGCVType == "complex"):
        raise TypeError(
            "`tangencyObstruction` requires its first argument to be a VFClass instance with DGCVType='complex'"
        )
    vf_real = allToReal(vf)
    real_eval = realPartOfVF(vf_real)(holToReal(graph_variable - CR_defining_expr))
    eval_real = allToReal(real_eval)
    substituted_expr = eval_real.subs(holToReal(symToReal(graph_variable)), symToReal(CR_defining_expr))
    result_expr = symToReal(substituted_expr)
    if simplify:
        result_expr = sp.simplify(result_expr)
    return result_expr


def weightedHomogeneousVF(
    arg1, arg2, arg3, arg4, degreeCap=0, _tempVar=None, assumeReal=None
):
    """
    Creates a weighted homogeneous vector field in a given coordinate space.

    This function generates a general weighted homogeneous vector field in the space of variables provided
    in *arg1*, with weights specified in *arg3*. The polynomial degree of variables with zero weight can
    be bounded by *degreeCap*.

    Parameters:
    -----------
    arg1 : tuple or list
        A tuple or list of variables, initialized by *varWithVF* or *complexVarProc*.
    arg2 : int
        An integer specifying the weight of the vector field.
    arg3 : list of int
        A list of non-negative integer weights corresponding to the variables in *arg1*.
    arg4 : str
        A string label for the variables in the returned vector field.
    degreeCap : int, optional
        Maximum polynomial degree for zero-weight variables (default is 0).
    _tempVar : any, optional
        Internal key.
    assumeReal : bool, optional
        Whether to assume the variables are real (default is False).

    Returns:
    --------
    VFClass
        A weighted homogeneous vector field.

    Raises:
    -------
    NA
    """
    pListLoc = []
    for j in range(len(arg3)):
        pListLoc.append(
            createPolynomial(
                arg4 + "_" + str(j) + "_",
                arg2 + arg3[j],
                arg1,
                degreeCap=degreeCap,
                weightedHomogeneity=arg3,
                _tempVar=_tempVar,
                assumeReal=assumeReal,
            )
        )
    return reduce(
        addVF,
        [
            scaleVF(pListLoc[j], eval("D_" + str(arg1[j]), _cached_caller_globals))
            for j in range(len(arg1))
        ],
    )


def findWeightedCRSymmetries_old(
    arg1,
    arg2,
    arg3,
    arg4,
    arg5,
    arg6,
    degreeCap=0,
    returnVectorFieldBasis=False,
    applyNumer=False,
    simplifyingFactor=None,
    simplify=False
):
    """
    ***This function's algorithm will be revised in future updates***

    Attempts to find all infinitesimal symmetries of a rigid CR hypersurface given by setting one variable *arg5* equal to a defining function *arg1* in the variable space *arg2* with weighted homogeneity *arg3* w.r.t. to non-negative integer weights in *arg4*. Variables in the returned vector field's coefficients are labeled by *arg6*.

    Only polynomial vector fields are searched for, so if a variable is assigned weight zero, the function cannot search across general symmetries. In such cases, it rather searches all possible symmetries with polynomial degree in the zero-weighted variables up to the specified bound *degreeCap*. If *degreeCap* is not specified, then it defaults to zero.

    The algorithm is most succesful when the function *arg1* must be a polynomial.

    The function *arg1* should not depend on the variable *arg5* (i.e., the algorithm is not intended for implicit defining equations). If there is such dependence *findWeightedCRSymmetries* may still find some but not all symmetries.


    Args:
        arg1: Defining function of a rigid CR hypersurface.
        arg2: a tuple or list of complex variables parameterizing the space that the above CR hypersurface is defined in (not including the transverse direction symmetry).
        arg3: list of non-negative integer weights corresponding to the variables in *arg1* (must have the same length as *arg1*). If 0 is among the weights, then then proceedure will only test polynomial vector fields with polynomial degree in the weight zero variables up to the weight specified in *degreeCap*. By default, degreeCap=0, and can be set to any positive integer.
        arg4: int
        arg5: The real variable that when set equal to the defining function defines the CR hypersurface.
        arg6: str
        degreeCap: (optional keyword) set this keyword argument equal to any positive integer. If not specified, it defaults to zero.
        applyNumer: (optional keyword) True or False. Set equal to true if defining equation is rational but not polynomial. It can help the internal solvers.

    Returns:
        coefficient list for a holomorphic vector field containing variables, and any set real value for these variables defines an infinitesimal symmetry. **Note, indeed only real values for the variables define actual symmetries**

    Raises:
        NA
    """
    def extractRIVar(arg1):
        return sum([list(holToReal(j).atoms(sp.Symbol)) for j in arg1], [])

    VFLoc = addVF(
        weightedHomogeneousVF(
            arg2,
            arg4,
            arg3,
            "ALoc",
            _tempVar=retrieve_passkey(),
            degreeCap=degreeCap,
            assumeReal=True,
        ),
        scaleVF(sp.I,weightedHomogeneousVF(
                arg2,
                arg4,
                arg3,
                "BLoc",
                _tempVar=retrieve_passkey(),
                degreeCap=degreeCap,
                assumeReal=True,
            ),
        ),
    )
    tanObst = tangencyObstruction(VFLoc, arg1, arg5, simplify=simplify)
    varLoc = tanObst.atoms(sp.Symbol)
    varLoc1 = {j for j in varLoc}
    varComp = set(extractRIVar(arg2))
    varLoc.difference_update(varComp)
    varLoc1.difference_update(varLoc)
    varComp.difference_update(varLoc1)
    variableProcedure(arg6, len(varLoc), assumeReal=True)
    if applyNumer:
        if varLoc1 == set():
            varLoc1 = set(arg2)
        coefList = sp.poly_from_expr(sp.expand(sp.numer(tanObst)), *varLoc1)[0].coeffs()
        solutions = sp.solve(coefList, varLoc)
    elif simplifyingFactor is None:
        if varLoc1 == set():
            varLoc1 = set(arg2)
        coefList = sp.poly_from_expr(sp.expand(tanObst), *varLoc1)[0].coeffs()
        solutions = sp.solve(coefList, varLoc)
    else:
        if varLoc1 == set():
            varLoc1 = set(arg2)

        coefList = sp.poly_from_expr(
            sp.expand(simplify(symToReal(simplifyingFactor) * tanObst)), *varLoc1
        )[0].coeffs()
        solutions = sp.solve(coefList, varLoc)
    if len(solutions) == 0:
        if tanObst!=0:
            clearVar(*listVar(temporary_only=True), report=False)
            raise ValueError(f"no solution to this system: {coefList}")
        else:
            solutions=dict()
    if type(solutions) is dict:
        VFCLoc = [j.subs(solutions) for j in holVF_coeffs(VFLoc, arg2)]
        subVar = sum(VFCLoc).atoms(sp.Symbol)
        subVar.difference_update(set(arg2))
        variableProcedure(arg6, len(subVar), assumeReal=True)
        VFCLoc = [
            j.subs(dict(zip(subVar, eval(arg6, _cached_caller_globals))))
            for j in VFCLoc
        ]
        clearVar(*listVar(temporary_only=True), report=False)
        if returnVectorFieldBasis:
            VFListLoc = []
            for j in eval(arg6, _cached_caller_globals):
                VFCLocTemp = [
                    k.subs(j, 1).subs(
                        [(ll, 0) for ll in eval(arg6, _cached_caller_globals)]
                    )
                    for k in VFCLoc
                ]
                VFListLoc.append(assembleFromHolVFC(VFCLocTemp, arg2))
            clearVar(arg6, report=False)
            return VFListLoc, VFCLoc
        else:
            return VFCLoc
    else:
        VFCLoc = holVF_coeffs(VFLoc, arg2)
        subVar = sum(VFCLoc).atoms(sp.Symbol)
        subVar.difference_update(set(arg2))
        variableProcedure(arg6, len(subVar), assumeReal=True)
        VFCLoc = [
            j.subs(dict(zip(subVar, eval(arg6, _cached_caller_globals))))
            for j in VFCLoc
        ]
        clearVar(*listVar(temporary_only=True), report=False)
        return VFCLoc, solutions


def findWeightedCRSymmetries(
    graph_function,
    holomorphic_coordinates,
    coordinate_weights,
    symmetry_weight,
    graph_variable,
    coeff_label = None,
    degreeCap = 0,
    returnVectorFieldBasis = False,
    returnAllformats = False,
    simplifyingFactor = None,
    assume_polynomial = False,
    simplify=False
):
    """
    """
    if returnAllformats:
        returnVectorFieldBasis = True
    vr = get_variable_registry()
    cd = vr['conversion_dictionaries']
    def extractRIVar(coordinate_list):
        return set(sum([[cd['real_part'].get(var,var),cd['im_part'].get(var,var)] for var in coordinate_list], []))
    VFLoc = addVF(
        weightedHomogeneousVF(
            holomorphic_coordinates,
            symmetry_weight,
            coordinate_weights,
            "ALoc",
            _tempVar=retrieve_passkey(),
            degreeCap=degreeCap,
            assumeReal=True,
        ),
        scaleVF(sp.I,weightedHomogeneousVF(
                holomorphic_coordinates,
                symmetry_weight,
                coordinate_weights,
                "BLoc",
                _tempVar=retrieve_passkey(),
                degreeCap=degreeCap,
                assumeReal=True,
            ),
        ),
    )
    tanObst = tangencyObstruction(VFLoc, graph_function, graph_variable, simplify=simplify)
    varLoc = tanObst.atoms(sp.Symbol)
    varLoc1 = varLoc.copy()
    varComp = extractRIVar(holomorphic_coordinates)
    varLoc.difference_update(varComp)
    varLoc1.difference_update(varLoc)
    varComp.difference_update(varLoc1)
    if coeff_label is None:
        coeff_label = create_key(prefix='X',key_length=5)
    variableProcedure(coeff_label, len(varLoc), assumeReal=True)
    if simplifyingFactor is not None:
        tanObst = sp.simplify(symToReal(simplifyingFactor)*tanObst)
    if not assume_polynomial:
        if varLoc1 == set():
            varLoc1 = set(holomorphic_coordinates)
        coefList = sp.poly_from_expr(sp.expand(sp.numer(tanObst)), *varLoc1)[0].coeffs()
        solutions = solve_DGCV(coefList, varLoc, method="auto")
    else:
        if varLoc1 == set():
            varLoc1 = set(holomorphic_coordinates)

        coefList = sp.poly_from_expr(
            sp.expand(tanObst), *varLoc1
        )[0].coeffs()
        solutions = solve_DGCV(coefList, varLoc, method="auto")
    if len(solutions) == 0:
        if tanObst!=0:
            clearVar(*listVar(temporary_only=True), report=False)
            raise ValueError(f"no solution to this system: {coefList}")
        else:
            solutions=dict()
            VFCLoc = [0 for _ in holomorphic_coordinates]
    else:
        if isinstance(solutions,(list,tuple)):
            solutions = solutions[0]
        VFCLoc = [j.subs(solutions) for j in holVF_coeffs(VFLoc, holomorphic_coordinates)]

    subVar = set()
    for term in VFCLoc:
        subVar |= term.atoms(sp.Symbol)
    subVar.difference_update(set(holomorphic_coordinates))
    variableProcedure(coeff_label, len(subVar), assumeReal=True)
    coeff_vars = _cached_caller_globals[coeff_label]
    VFCLoc = [
        j.subs(dict(zip(subVar, coeff_vars)))
        for j in VFCLoc
    ]
    clearVar(*listVar(temporary_only=True), report=False)
    if returnVectorFieldBasis:
        VFListLoc = []
        for j in coeff_vars:
            VFCLocTemp = [
                k.subs(j, 1).subs(
                    [(ll, 0) for ll in coeff_vars]
                )
                for k in VFCLoc
            ]
            VFListLoc.append(assembleFromHolVFC(VFCLocTemp, holomorphic_coordinates))
        clearVar(coeff_label, report=False)
        if returnAllformats:
            return VFCLoc,VFListLoc
        return VFListLoc
    else:
        return VFCLoc




def model2Nondegenerate(arg1, arg2, arg3, arg4, return_matrices=False, simplify=True):
    """
    Builds the defining equation for a 2-nondegnerate model hypersurface using the general formula from the arXiv preprint arXiv:2404.06525.

    Args:
        arg1: nondegenerate s-by-s hermitian matrix
        arg2: s-by-s symmetric matrix valued function of some complex variables whose differential at zero is injective, and whose value at zero is zero.
        arg3: a length s tuple of complex variables, different from those appearing in *arg2*
        arg4: a single complex variable, different from those appearing in *arg2* and *arg3*

    Returns:
        A sympy expression. Setting this expression equal to zero defines the 2-nondegenerate model.

    Raises:
        NA
    """

    def format_mat(mat):
        if isinstance(mat,(tuple,list)):
            if all(isinstance(elem,(tuple,list)) for elem in mat):
                if len(set(len(elem) for elem in mat))==1:
                    mat = sp.Matrix(mat)
        if isinstance(mat,sp.Matrix):
            return mat
        else:
            raise TypeError('`model2Nondegenerate` expects first to arguments to be array-like data.')

    arg1 = format_mat(arg1)
    arg2 = format_mat(arg2)
    def simplify_rules(expr):
        if simplify:
            return sp.simplify(expr)
        else:
            return expr
    BARSLoc = sp.conjugate(arg2)
    zVecLoc = sp.Matrix(arg3)
    bzVecLoc = sp.Matrix([sp.conjugate(j) for j in arg3])
    sizeLoc = arg1.shape[0]
    hFun = (sp.Rational(1, 2)) * (
        (arg1 * (sp.eye(sizeLoc) - (BARSLoc * sp.Transpose(arg1) * arg2 * arg1)) ** (-1))
        + ((sp.eye(sizeLoc) - (arg1 * BARSLoc * sp.Transpose(arg1) * arg2)) ** (-1) * arg1)
    )
    sFun = (
        arg1
        * ((sp.eye(sizeLoc) - (BARSLoc * sp.Transpose(arg1) * arg2 * arg1)) ** (-1))
        * BARSLoc
        * sp.Transpose(arg1)
    )
    bsFun = (
        sp.Transpose(arg1)
        * ((sp.eye(sizeLoc) - (arg2 * arg1 * BARSLoc * sp.Transpose(arg1))) ** (-1))
        * arg2
        * arg1
    )
    if return_matrices:
        return (
            simplify_rules(
                (
                    sp.Transpose(zVecLoc) * hFun * bzVecLoc
                    + (sp.Rational(1, 2))
                    * (
                        sp.Transpose(zVecLoc) * sFun * zVecLoc
                        + sp.Transpose(bzVecLoc) * bsFun * bzVecLoc
                    )
                )[0]
            )
            - sp.im(arg4),
            hFun,
            sFun,
        )
    else:
        return simplify_rules(
            (
                sp.Transpose(zVecLoc) * hFun * bzVecLoc
                + (sp.Rational(1, 2))
                * (
                    sp.Transpose(zVecLoc) * sFun * zVecLoc
                    + sp.Transpose(bzVecLoc) * bsFun * bzVecLoc
                )
            )[0]
        ) - sp.im(arg4)



import sympy as sp

from ._safeguards import get_variable_registry


def _holToReal(expr, skipVar=None, simplify_everything=True):
    """
    Converts holomorphic variables from VMF in SymPy expression to real variables.

    Parameters:
    expr : sympy.Expr
        The expression to convert.
    skipVar : list of str, optional
        A list of complex variable system labels to skip during conversion.
        For any variable in skipVar, the associated real and imaginary variables
        will not be substituted.

    Returns:
    sympy.Expr
    """
    variable_registry = get_variable_registry()
    conversion_dict = (
        variable_registry.get("conversion_dictionaries", {}).get("holToReal", {}).copy()
    )
    def format(expr):
        return sp.sympify(expr).subs(conversion_dict)

    # Skip specified holomorphic variable systems if skipVar is provided
    if skipVar:
        for var in skipVar:
            complex_system = variable_registry.get("complex_variable_systems", {}).get(
                var, {}
            )
            family_values = complex_system.get("family_values", ((), (), (), ()))
            if complex_system.get("family_type", None) == "single":
                family_values = tuple([(j,) for j in family_values])

            # Remove the holomorphic variables (first tuple) from the conversion_dict
            holomorphic_vars = family_values[0]
            for hol_var in holomorphic_vars:
                if hol_var in conversion_dict:
                    del conversion_dict[hol_var]


    if hasattr(expr, "applyfunc"):
        return expr.applyfunc(format)
    else:
        return format(expr)


def _realToSym(expr, skipVar=None, simplify_everything=True):
    """
    Converts real variables from VMF in SymPy expression to symbolic conjugates.

    Parameters:
    expr : sympy.Expr
        The expression to convert.
    skipVar : list of str, optional
        A list of complex variable system labels to skip during conversion.
        For any variable in skipVar, the associated real and imaginary variables
        will not be substituted.

    Returns:
    sympy.Expr
    """
    variable_registry = get_variable_registry()
    conversion_dict = (
        variable_registry.get("conversion_dictionaries", {}).get("realToSym", {}).copy()
    )

    def format(expr):
        return sp.sympify(expr).subs(conversion_dict)

    # Skip specified real and imaginary variable systems if skipVar is provided
    if skipVar:
        for var in skipVar:
            complex_system = variable_registry.get("complex_variable_systems", {}).get(
                var, {}
            )
            family_values = complex_system.get("family_values", ((), (), (), ()))
            if complex_system.get("family_type", None) == "single":
                family_values = tuple([(j,) for j in family_values])

            # Remove both the real (third tuple) and imaginary (fourth tuple) variables from the conversion_dict
            real_vars = family_values[2] + family_values[3]
            for real_var in real_vars:
                if real_var in conversion_dict:
                    del conversion_dict[real_var]

    if hasattr(expr, "applyfunc"):
        return expr.applyfunc(format)
    else:
        return format(expr)


def _symToHol(expr, skipVar=None, simplify_everything=True):
    """
    Converts symbolic conjugated variables from VMF in SymPy expression to holomorphic variables.

    Parameters:
    expr : sympy.Expr
        The expression to convert.
    skipVar : list of str, optional
        A list of complex variable system labels to skip during conversion.
        For any variable in skipVar, the associated real and imaginary variables
        will not be substituted.

    Returns:
    sympy.Expr
    """
    variable_registry = get_variable_registry()
    conversion_dict = (
        variable_registry.get("conversion_dictionaries", {}).get("symToHol", {}).copy()
    )

    def format(expr):
        return sp.sympify(expr).subs(conversion_dict)

    # If skipVar is provided, modify the conversion_dict to exclude specified variables.
    if skipVar:
        for var in skipVar:
            # Access the complex variable system for the skipped variable
            complex_system = variable_registry.get("complex_variable_systems", {}).get(
                var, {}
            )
            family_values = complex_system.get("family_values", ((), (), (), ()))
            if complex_system.get("family_type", None) == "single":
                family_values = tuple([(j,) for j in family_values])

            # The second tuple contains the antiholomorphic variables
            antiholomorphic_vars = family_values[1]

            # Remove the antiholomorphic variables from the conversion_dict
            for anti_var in antiholomorphic_vars:
                if anti_var in conversion_dict:
                    del conversion_dict[anti_var]

    if hasattr(expr, "applyfunc"):
        return expr.applyfunc(format)
    else:
        return format(expr)


def _holToSym(expr, skipVar=None, simplify_everything=True):
    """
    Converts holomorphic variables from VMF in an expression to symbolic 
    conjugates. This is done by first converting holomorphic variables to real 
    variables, and then converting real variables to symbolic conjugates.

    Note: This process will also convert any present real variables from the 
    VMF (both real and imaginary parts) to their symbolic conjugate format.

    Parameters:
    expr : sympy.Expr
        The expression to convert.
    skipVar : list of str, optional
        A list of complex variable system labels to skip during conversion.
        For any variable in skipVar, the associated real and imaginary variables
        will not be substituted.

    Returns:
    sympy.Expr
    """
    # First apply holToReal()
    expr = _holToReal(expr, skipVar=skipVar, simplify_everything=simplify_everything)

    # Then apply realToSym()
    expr = _realToSym(expr, skipVar=skipVar, simplify_everything=simplify_everything)

    return expr


def _realToHol(expr, skipVar=None, simplify_everything=True):
    """
    Converts real variables from VMF in SymPy expression to holomorphic variables.

    Parameters:
    expr : sympy.Expr
        The expression to convert.
    skipVar : list of str, optional
        A list of complex variable system labels to skip during conversion.
        For any variable in skipVar, the associated real and imaginary variables
        will not be substituted.

    Returns:
    sympy.Expr
    """
    variable_registry = get_variable_registry()
    conversion_dict = (
        variable_registry.get("conversion_dictionaries", {}).get("realToHol", {}).copy()
    )

    def format(expr):
        return sp.sympify(expr).subs(conversion_dict)

    # Skip specified real and imaginary variable systems if skipVar is provided
    if skipVar:
        for var in skipVar:
            complex_system = variable_registry.get("complex_variable_systems", {}).get(
                var, {}
            )
            family_values = complex_system.get("family_values", ((), (), (), ()))
            if complex_system.get("family_type", None) == "single":
                family_values = tuple([(j,) for j in family_values])

            # Remove both the real (third tuple) and imaginary (fourth tuple) variables from the conversion_dict
            real_vars = family_values[2] + family_values[3]
            for real_var in real_vars:
                if real_var in conversion_dict:
                    del conversion_dict[real_var]

    if hasattr(expr, "applyfunc"):
        return expr.applyfunc(format)
    else:
        return format(expr)


def _symToReal(expr, skipVar=None, simplify_everything=True):
    """
    Converts symbolic conjugates from VMF in SymPy expression to real variables.

    Parameters:
    expr : sympy.Expr
        The expression to convert.
    skipVar : list of str, optional
        A list of complex variable system labels to skip during conversion.
        For any variable in skipVar, the associated real and imaginary variables
        will not be substituted.

    Returns:
    sympy.Expr
    """
    variable_registry = get_variable_registry()
    conversion_dict = (
        variable_registry.get("conversion_dictionaries", {}).get("symToReal", {}).copy()
    )

    def format(expr):
        return sp.sympify(expr).subs(conversion_dict)

    # Skip specified symbolic conjugate systems if skipVar is provided
    if skipVar:
        for var in skipVar:
            complex_system = variable_registry.get("complex_variable_systems", {}).get(
                var, {}
            )
            family_values = complex_system.get("family_values", ((), (), (), ()))
            if complex_system.get("family_type", None) == "single":
                family_values = tuple([(j,) for j in family_values])

            # Remove the symbolic conjugates (second tuple) from the conversion_dict
            antiholomorphic_vars = family_values[1]
            for anti_var in antiholomorphic_vars:
                if anti_var in conversion_dict:
                    del conversion_dict[anti_var]

    if hasattr(expr, "applyfunc"):
        return expr.applyfunc(format)
    else:
        return format(expr)


def _allToReal(expr, skipVar=None, simplify_everything=True):
    """
    Converts all variables in VMF (holomorphic, symbolic conjugates, and real)
    to real variables.

    Parameters:
    expr : sympy.Expr
        The expression to convert.
    skipVar : list of str, optional
        A list of variable system labels to skip during conversion.

    Returns:
    sympy.Expr
    """
    # First convert symbolic conjugates to real variables
    expr = _symToReal(expr, skipVar=skipVar, simplify_everything=simplify_everything)

    # Then convert holomorphic variables to real variables
    expr = _holToReal(expr, skipVar=skipVar, simplify_everything=simplify_everything)

    return expr


def _allToHol(expr, skipVar=None, simplify_everything=True):
    """
    Converts all variables in VMF (real, symbolic conjugates, and holomorphic)
    to holomorphic variables.

    Parameters:
    expr : sympy.Expr
        The expression to convert.
    skipVar : list of str, optional
        A list of variable system labels to skip during conversion.

    Returns:
    sympy.Expr
    """

    if isinstance(expr, list):
        return [_allToHol(j) for j in expr]
    if isinstance(expr, tuple):
        return [_allToHol(j) for j in expr]

    # First convert symbolic conjugates to real variables
    expr = _symToHol(expr, skipVar=skipVar, simplify_everything=simplify_everything)

    # Then convert real variables to holomorphic variables
    expr = _realToHol(expr, skipVar=skipVar, simplify_everything=simplify_everything)

    return expr


def _allToSym(expr, skipVar=None, simplify_everything=True):
    """
    Converts all variables from VMF (holomorphic, real, and symbolic conjugates) to
    symbolic conjugates.

    Parameters:
    expr : sympy.Expr
        The expression to convert.
    skipVar : list of str, optional
        A list of variable system labels to skip during conversion.

    Returns:
    sympy.Expr
    """
    return _holToSym(expr, skipVar=skipVar, simplify_everything=simplify_everything)




from setuptools import find_packages, setup

long_description = """
# dgcv - Differential Geometry with Complex Variables

dgcv integrates tools for differential geometry with systematic handling of complex variables-related structures.

## Tutorials

To get started, check out the Jupyter Notebook tutorials:

- **[dgcv Introduction](https://www.realandimaginary.com/dgcv/tutorials/DGCV_introduction/)**: An introduction to the key concepts and setup.
- **[dgcv in Action](https://www.realandimaginary.com/dgcv/tutorials/DGCV_in_action/)**: A quick tour through examples from some of the library's more elaborate functions.
"""

setup(
    name="dgcv",
    version="0.2.15",
    description="Differential Geometry with Complex Variables",
    long_description=long_description,  # This shows up on PyPI
    long_description_content_type="text/markdown",
    package_dir={"": "src"},  # This tells setuptools that packages are under src/
    packages=find_packages(where="src"),
    package_data={
        "dgcv": ["assets/fonts/*.ttf", "assets/fonts/fonts.css"],  # Include font files
    },
    include_package_data=True,
    python_requires=">=3.6",
    install_requires=["sympy>=1.9", "pandas>=1.0", "ipython>=7.0"],
)



import re

import sympy as sp

from .config import greek_letters

joinders = {'symmetric':{'latex':'\\odot ','plain':'&'},'skew':{'latex':'\\wedge ','plain':'*'}}
def _shape_joinders(shape,format):
    if shape in ['symmetric','skew']:
        return joinders[shape][format]
    if format == 'latex':
        return '\\otimes '
    return '@'

def tensor_field_printer(tensor):
    terms = tensor.coeff_dict
    valence = tensor.valence

    def coeff_formatter(scalar):
        """
        Formats scalar coefficients for printing.
        """
        if scalar == 1:
            return ""
        if scalar == -1:
            return "-"
        scalar_str = str(scalar)
        if isinstance(scalar, sp.Basic) and len(scalar.args) > 0:
            return f"({scalar_str})*"
        return scalar_str+'*'

    formatted_terms = []
    for vec, scalar in terms.items():
        if scalar != 0:
            basis_elements = [
                f"D_{tensor.varSpace[vec[j]]}" if valence[j] == 1 else f"d_{tensor.varSpace[vec[j]]}"
                for j in range(len(valence))
            ]
            basis_element = _shape_joinders(tensor.data_shape,'plain').join(basis_elements)
            formatted_terms.append(f"{coeff_formatter(scalar)}{basis_element}")

    if not formatted_terms:  # If all scalars are zero
        pref_index = tensor._simplifyKW['preferred_basis_element'] if tensor._simplifyKW['preferred_basis_element'] else (0,) * len(valence)
        basis_elements = [
            f"D_{tensor.varSpace[pref_index[j]]}" if valence[j] == 1 else f"d_{tensor.varSpace[pref_index[j]]}"
            for j in range(len(valence))
        ]
        formatted_terms.append(f"0{_shape_joinders(tensor.data_shape,'plain').join(basis_elements)}")

    # join terms
    result = formatted_terms[0]
    for term in formatted_terms[1:]:
        if term.startswith("-"):
            result += term
        else:
            result += f"+{term}"

    return result

def tensor_field_latex(tensor):
    """
    Generates LaTeX representation of a tensorField based on its data shape.
    """
    terms = tensor.coeff_dict
    valence = tensor.valence

    def coeff_formatter(scalar):
        """
        Formats scalar coefficients for LaTeX.

        - Wraps scalar expressions in parentheses if they contain multiple terms.
        - Handles special cases like 1 and -1.
        """
        if scalar == 1:
            return ""
        if scalar == -1:
            return "-"

        scalar_str = sp.latex(scalar)

        # Use is_Add to check for multi-term expressions
        if sp.sympify(scalar).is_Add:
            return f"\\left({scalar_str}\\right)"

        return scalar_str

    formatted_terms = []

    for vec, scalar in terms.items():
        if scalar != 0:
            basis_elements = [
                f"\\frac{{\\partial}}{{\\partial {_process_var_label(tensor.varSpace[index])}}}"
                if valence[j] == 1
                else f"d {_process_var_label(tensor.varSpace[index])}"
                for j, index in enumerate(vec)
            ]
            basis_element = _shape_joinders(tensor.data_shape, 'latex').join(basis_elements)
            formatted_terms.append(f"{coeff_formatter(scalar)} {basis_element}")

    # Handle case where all coefficients are zero
    if not terms or all(scalar == 0 for scalar in terms.values()):
        pref_index = tensor._simplifyKW['preferred_basis_element'] if tensor._simplifyKW['preferred_basis_element'] else (0,) * len(valence)
        basis_elements = [
            f"\\frac{{\\partial}}{{\\partial {_process_var_label(tensor.varSpace[pref_index[j]])}}}"
            if valence[j] == 1
            else f"d {_process_var_label(tensor.varSpace[pref_index[j]])}"
            for j in range(len(valence))
        ]
        formatted_terms.append("0" + _shape_joinders(tensor.data_shape, 'latex').join(basis_elements))

    # Join terms with "+" and handle "+-" cases
    latex_str = " + ".join(formatted_terms).replace("+ -", "- ")

    return f"${latex_str}$"


def tensor_VS_printer(tp):
    terms = tp.coeff_dict

    def coeff_formatter(scalar):
        """
        Formats scalar coefficients for printing.
        """
        if scalar == 1:
            return ""
        if scalar == -1:
            return "-"
        scalar_str = str(scalar)
        if isinstance(scalar, sp.Basic) and len(scalar.args) > 0:
            return f"({scalar_str})*"
        return scalar_str+'*'

    basis_labels = tp.vector_space.basis_labels or [f"e_{i+1}" for i in range(tp.vector_space.dimension)]

    formatted_terms = []
    for vec, scalar in terms.items():
        valence = vec[len(vec)//2:]
        vec = vec[:len(vec)//2]
        if scalar != 0:
            basis_elements = [
                f"{_process_var_label(basis_labels[vec[j]])}"
                if valence[j] == 1
                else f"{_process_var_label(basis_labels[vec[j]])}^''"
                for j in range(len(valence))
            ]
            basis_element = '@'.join(basis_elements)
            formatted_terms.append(f"{coeff_formatter(scalar)}{basis_element}")

    if not formatted_terms:  # If all scalars are zero
        basis_elements = [
            f"{_process_var_label(basis_labels[vec[j]])}"
            for j in range(tp.max_degree)
        ]
        formatted_terms.append(f"0{'@'.join(basis_elements)}")

    # join terms
    result = formatted_terms[0]
    for term in formatted_terms[1:]:
        if term.startswith("-"):
            result += term
        else:
            result += f"+{term}"

    return result

def tensor_VS_latex(tp):
    """
    Generates LaTeX representation of a tensorProduct.
    """
    terms = tp.coeff_dict

    def coeff_formatter(scalar):
        """
        Formats scalar coefficients for LaTeX.

        Wraps scalar expressions in parentheses if they have more than one term.
        """
        if scalar==1:
            return ""
        if scalar==-1:
            return "-"
        scalar_str = sp.latex(scalar)
        if isinstance(scalar, sp.Expr) and len(scalar.args) > 0:
            return f"\\left({scalar_str}\\right)"
        return scalar_str
    basis_labels = tp.vector_space.basis_labels or [f"e_{i+1}" for i in range(tp.vector_space.dimension)]
    formatted_terms = []
    for vec, scalar in terms.items():
        valence = vec[len(vec)//2:]
        vec = vec[:len(vec)//2]
        if scalar != 0:
            basis_elements = [
                f"{_process_var_label(basis_labels[vec[j]])}"
                if valence[j] == 1
                else f"{_process_var_label(basis_labels[vec[j]])}^*"
                for j in range(len(valence))
            ]
            basis_element = '\\otimes '.join(basis_elements)
            formatted_terms.append(f"{coeff_formatter(scalar)} {basis_element}")

    if not formatted_terms:  # If all scalars are zero
        basis_elements = [
            f"{_process_var_label(basis_labels[vec[j]])}"
            for j in range(tp.max_degree)
        ]
        formatted_terms.append("0" + '\\otimes '.join(basis_elements))

    # Join terms
    latex_str = formatted_terms[0]
    for term in formatted_terms[1:]:
        if term.startswith("-"):
            latex_str += term
        else:
            latex_str += f" + {term}"

    return f"${latex_str}$"

def tensor_latex_helper(tp):
    """
    Generates LaTeX representation of a tensorProduct for SymPy's _latex method.
    """
    terms = tp.coeff_dict

    def coeff_formatter(scalar):
        """
        Formats scalar coefficients for LaTeX.

        Wraps scalar expressions in parentheses if they have more than one term.
        """
        if scalar == 1:
            return ""
        if scalar == -1:
            return "-"
        scalar_str = sp.latex(scalar)
        if isinstance(scalar, sp.Expr) and len(scalar.args) > 0:
            return f"\\left({scalar_str}\\right)"
        return scalar_str

    basis_labels = tp.vector_space.basis_labels or [f"e_{i+1}" for i in range(tp.vector_space.dimension)]
    formatted_terms = []
    for vec, scalar in terms.items():
        valence = vec[len(vec) // 2:]
        vec = vec[:len(vec) // 2]
        if scalar != 0:
            basis_elements = [
                f"{_process_var_label(basis_labels[vec[j]])}"
                if valence[j] == 1
                else f"{_process_var_label(basis_labels[vec[j]])}^*"
                for j in range(len(valence))
            ]
            basis_element = '\\otimes '.join(basis_elements)
            formatted_terms.append(rf"{coeff_formatter(scalar)} {basis_element}")

    if not formatted_terms:  # If all scalars are zero
        basis_elements = [
            f"{_process_var_label(basis_labels[vec[j]])}"
            for j in range(tp.max_degree)
        ]
        formatted_terms.append("0" + '\\otimes '.join(basis_elements))

    # Join terms
    latex_str = formatted_terms[0]
    for term in formatted_terms[1:]:
        if term.startswith("-"):
            latex_str += term
        else:
            latex_str += f" + {term}"

    return latex_str




def _convert_to_greek(var_name):
    for name, greek in greek_letters.items():
        if var_name.startswith(name):
            return var_name.replace(name, greek, 1)
    return var_name

def _process_var_label(var):
    """
    Converts variable names into LaTeX-compatible labels.

    - Recognizes Greek letter names and converts them to proper LaTeX.
    - If the variable has a numerical subscript (e.g., "theta1"), it formats as "\theta_1".
    - Handles "BAR" prefix for conjugates, wrapping the final output in \\overline{}.
    """
    var_str = str(var)
    is_conjugate = False

    # Handle "BAR" prefix
    if var_str.startswith("BAR"):
        var_str = var_str[3:]  # Remove "BAR"
        is_conjugate = True

    # Match label + optional trailing numbers
    match = re.match(r"(.*?)(\d*)$", var_str)

    if match:
        label_part = match.group(1).rstrip("_")  # Remove trailing underscores
        number_part = match.group(2)  # Extract number if present

        # Special latex symbols filter
        label_part = _convert_to_greek(label_part)

        # Format final output
        formatted_label = f"{label_part}_{{{number_part}}}" if number_part else label_part

        # Wrap in \overline{} if originally conjugate
        return f"\\overline{{{formatted_label}}}" if is_conjugate else formatted_label

    return var_str  # If regex fails, return unchanged



"""
DGCV: Differential Geometry with Complex Variables

This module provides tools uniquely relevant for complex differential geometry within the DGCV package. It includes Dolbeault operators (Del and DelBar) and a class for constructing and analyzing Kähler structures.

Key Functions:
    - Del(): Applies the holomorphic Dolbeault operator ∂ to a differential form or scalar.
    - DelBar(): Applies the antiholomorphic Dolbeault operator ∂̅ to a differential form or scalar.

Key Classes:
    - KahlerStructure: Represents a Kähler structure, with properties and attributes to compute many of their invariants. 

Author: David Sykes (https://github.com/YikesItsSykes)

Dependencies:
    - sympy

License:
    MIT License
"""

############## dependencies

import warnings

import sympy as sp

from .config import get_variable_registry
from .dgcv_core import (
    DFClass,
    STFClass,
    addDF,
    allToSym,
    changeDFBasis,
    complex_struct_op,
    realToSym,
    symToReal,
    tensorField,
)
from .Riemannian_geometry import metricClass
from .vector_fields_and_differential_forms import exteriorDerivative, makeZeroForm

############## Dolbeault operators


def Del(arg1):
    """
    Applies the holomorphic Dolbeault operator ∂ (Del) to a differential form or scalar.

    The Del operator takes a differential form or scalar and returns the result of applying the
    exterior derivative with respect to the holomorphic coordinates.

    Parameters:
    -----------
    arg1 : DFClass or sympy.Expr
        A differential form (DFClass) with DGCVType='complex', or a scalar (interpreted as a zero-form).

    Returns:
    --------
    DFClass
        The result of applying the Del operator, which is a differential form.

    Raises:
    -------
    TypeError
        If the input is not a DFClass or scalar (sympy.Expr), or if the DFClass has DGCVType='standard'.
    """
    variable_registry = get_variable_registry()
    # Ensure arg1 is a DFClass or convert it into a zero-form
    if not isinstance(arg1, DFClass) and isinstance(arg1, (int, float, sp.Expr)):
        arg1 = allToSym(arg1)
        varSpace = tuple(
            [
                j
                for j in sp.sympify(arg1).free_symbols
                if variable_registry["conversion_dictionaries"]["symToReal"]
            ]
        )
        intermediateDF = makeZeroForm(
            arg1, varSpace=varSpace, DGCVType="complex", default_var_format="complex"
        )
        arg1 = makeZeroForm(
            arg1,
            intermediateDF.compVarSpace,
            DGCVType="complex",
            default_var_format="complex",
        )

    elif isinstance(arg1, DFClass):
        if arg1.DGCVType == "standard":
            raise TypeError(
                "`Del` only operates on DFClass differential forms with DGCVType='complex' (and scalars like sympy.Expr, which it interprets as 0-forms). Tip: set 'complex=True' in the `createVariables` variable creation function to initialize a coframe with `DGCVType='complex'.`"
            )
        else:
            arg1 = realToSym(symToReal(arg1))
    else:
        raise TypeError(
            "`Del` only operate on differential forms, DFClass (and scalars like sympy.Expr, which it interprets as 0-forms)."
        )

    # Helper function to compute the exterior derivative of a zero-form
    def DelOfZeroForm(arg1):
        HVSpace = arg1.holVarSpace
        sparseDataLoc = {
            (j,): sp.diff(allToSym(arg1.coeffsInKFormBasis[0]), HVSpace[j])
            for j in range(len(HVSpace))
        }
        return DFClass(arg1.varSpace, sparseDataLoc, 1, DGCVType="complex")

    # Handle zero-forms
    if arg1.degree == 0:
        return DelOfZeroForm(arg1)

    # Handle higher-degree forms
    minDataLoc = arg1.DFClassDataMinimal
    coeffsTo1Forms = [
        DelOfZeroForm(makeZeroForm(j[1], varSpace=arg1.varSpace, DGCVType="complex"))
        for j in minDataLoc
    ]
    # Construct the corresponding basis k-forms
    basisOfCoeffs = [
        DFClass(arg1.varSpace, {tuple(j[0]): 1}, arg1.degree, DGCVType="complex")
        for j in minDataLoc
    ]
    # Multiply the one-forms by the basis and sum them
    return addDF(
        *[coeffsTo1Forms[j] * basisOfCoeffs[j] for j in range(len(minDataLoc))]
    )


def DelBar(arg1):
    """
    Applies the antiholomorphic Dolbeault operator ∂̅ (DelBar) to a differential form or scalar.

    The DelBar operator takes a differential form or scalar and returns the result of applying the
    exterior derivative with respect to the antiholomorphic coordinates.

    Parameters:
    -----------
    arg1 : DFClass or sympy.Expr
        A differential form (DFClass) with DGCVType='complex', or a scalar (interpreted as a zero-form).

    Returns:
    --------
    DFClass
        The result of applying the DelBar operator, which is a differential form.

    Raises:
    -------
    TypeError
        If the input is not a DFClass or scalar (sympy.Expr), or if the DFClass has DGCVType='standard'.
    """
    variable_registry = get_variable_registry()
    # Ensure arg1 is a DFClass or convert it into a zero-form
    if not isinstance(arg1, DFClass) and isinstance(arg1, (int, float, sp.Expr)):
        arg1 = allToSym(arg1)
        varSpace = tuple(
            [
                j
                for j in sp.sympify(arg1).free_symbols
                if variable_registry["conversion_dictionaries"]["symToReal"]
            ]
        )
        intermediateDF = makeZeroForm(
            arg1, varSpace=varSpace, DGCVType="complex", default_var_format="complex"
        )
        arg1 = makeZeroForm(
            arg1,
            intermediateDF.compVarSpace,
            DGCVType="complex",
            default_var_format="complex",
        )

    if isinstance(arg1, DFClass):
        if arg1.DGCVType == "standard":
            raise TypeError(
                "`DelBar` only operates on DFClass differential forms with DGCVType='complex' (and scalars like sympy.Expr, which it interprets as 0-forms). Tip: set 'complex=True' in the `createVariables` variable creation function to initialize a coframe with `DGCVType='complex'.`"
            )
        else:
            arg1 = allToSym(symToReal(arg1))
    else:
        raise TypeError(
            "`DelBar` only operate on differential forms, DFClass (and scalars like sympy.Expr, which it interprets as 0-forms)."
        )

    # Helper function to compute the exterior derivative of a zero-form
    def DelBarOfZeroForm(arg1):
        AHVspace = arg1.antiholVarSpace
        CDim = len(AHVspace)
        sparseDataLoc = {
            (j + CDim,): sp.diff(allToSym(arg1.coeffsInKFormBasis[0]), AHVspace[j])
            for j in range(CDim)
        }
        return DFClass(arg1.varSpace, sparseDataLoc, 1, DGCVType="complex")

    # Handle zero-forms
    if arg1.degree == 0:
        return DelBarOfZeroForm(arg1)

    # Handle higher-degree forms
    minDataLoc = arg1.DFClassDataMinimal
    coeffsTo1Forms = [
        DelBarOfZeroForm(makeZeroForm(j[1], varSpace=arg1.varSpace, DGCVType="complex"))
        for j in minDataLoc
    ]
    # Construct the corresponding basis k-forms
    basisOfCoeffs = [
        DFClass(arg1.varSpace, {tuple(j[0]): 1}, arg1.degree, DGCVType="complex")
        for j in minDataLoc
    ]

    # Multiply the one-forms by the basis and sum them
    return addDF(
        *[coeffsTo1Forms[j] * basisOfCoeffs[j] for j in range(len(minDataLoc))]
    )


############## Kahler geometry


class KahlerStructure(sp.Basic):
    r"""
    Represents a Kähler structure, including its metric, , and Bochner tensor.

    The Kähler structure is defined by a symplectic form (Kähler form) and a coordinate space. The class
    provides methods to compute various geometric objects associated with Kähler manifolds, including the
    its metric, Holomorphic Riemann and Ricci Curvatures, and the Bochner tensor.

    Parameters:
    -----------
    varSpace : tuple of sympy.Symbol
        A tuple of variables from DGCV's complex coordinate systems representing the coordinate space of the manifold.
    kahlerForm : DFClass
        A Kähler form representing the symplectic structure.

    Attributes:
    -----------
    metric : metricClass
        The Riemannian metric associated with the Kähler structure.
    holRiemann : TFClass
        The (0,4) Riemann curvature tensor with the complex structure operator hooked into it second and fourth positions.
    holRicci : TFClass
        The Ricci curvature tensor with the complex structure operator hooked into its second position
    Bochner : TFClass
        The Bochner tensor in \( T^{1,0}M \otimes T^{0,1}M \otimes T^{1,0}M \otimes T^{0,1}M \).

    Raises:
    -------
    TypeError
        If the variable space includes a mixture of real and holomorphic coordinates.
    """

    def __new__(cls, varSpace, kahlerForm):
        # Call Basic.__new__ with only the positional arguments
        obj = sp.Basic.__new__(cls, varSpace, kahlerForm)
        return obj

    def __init__(self, varSpace, kahlerForm):
        variable_registry = get_variable_registry()
        if all(
            var in variable_registry["conversion_dictionaries"]["realToSym"]
            for var in varSpace
        ):
            self._varSpace_type = "real"
        elif all(
            var in variable_registry["conversion_dictionaries"]["symToReal"]
            for var in varSpace
        ):
            self._varSpace_type = "complex"
        else:
            raise TypeError(
                "The variable space given to the Kahler structure initialer has a mixture of real and holomorphic coordinates, which is not a supported data format."
            )
        variableSpaces = KahlerStructure.validate(varSpace)
        self.realVarSpace = variableSpaces[0]
        self.compVarSpace = variableSpaces[1] + variableSpaces[2]
        self.holVarSpace = variableSpaces[1]
        self.antiholVarSpace = variableSpaces[2]
        if self._varSpace_type == "real":
            self.varSpace = self.realVarSpace
        else:
            self.varSpace = self.compVarSpace
        self.kahlerForm = changeDFBasis(kahlerForm, self.varSpace)
        VFBasis = []
        for var in self.varSpace:
            varStr = str(var)
            for parent in variable_registry["complex_variable_systems"]:
                if (
                    varStr
                    in variable_registry["complex_variable_systems"][parent][
                        "variable_relatives"
                    ]
                ):
                    VFBasis = VFBasis + [
                        variable_registry["complex_variable_systems"][parent][
                            "variable_relatives"
                        ][varStr]["VFClass"]
                    ]
        self.coor_frame = VFBasis
        if self._varSpace_type == "real":
            self.coor_frame_real = self.coor_frame
            VFBasisHol = []
            for var in self.compVarSpace:
                varStr = str(var)
                for parent in variable_registry["complex_variable_systems"]:
                    if (
                        varStr
                        in variable_registry["complex_variable_systems"][parent][
                            "variable_relatives"
                        ]
                    ):
                        VFBasisHol = VFBasisHol + [
                            variable_registry["complex_variable_systems"][parent][
                                "variable_relatives"
                            ][varStr]["VFClass"]
                        ]
            self.coor_frame_complex = VFBasisHol
        else:
            self.coor_frame_complex = self.coor_frame
            VFBasisReal = []
            for var in self.compVarSpace:
                varStr = str(var)
                for parent in variable_registry["complex_variable_systems"]:
                    if (
                        varStr
                        in variable_registry["complex_variable_systems"][parent][
                            "variable_relatives"
                        ]
                    ):
                        VFBasisReal = VFBasisReal + [
                            variable_registry["complex_variable_systems"][parent][
                                "variable_relatives"
                            ][varStr]["VFClass"]
                        ]
            self.coor_frame_real = VFBasisReal

        self._metric = None
        self._is_closed = None
        self._holRiemann = None  # holomorphic Riemann tensor
        self._holRicci = None  # holomorphic Ricci tensor
        self._Bochner = None  # Bochner tensor

    @staticmethod
    def validate(varSpace):  # validates and organizes given varspace
        """
        Validates and organizes a given variable space into real, holomorphic, and antiholomorphic components.

        This method checks the input variable space and categorizes the variables into real, holomorphic,
        and antiholomorphic components, based on the relationships between the variables.

        Parameters:
        -----------
        varSpace : tuple of sympy.Symbol
            A tuple of complex variables representing the coordinate space.

        Returns:
        --------
        list of tuples
            A list containing the real and imaginary variable spaces, holomorphic variable space, and
            antiholomorphic variable space.
        """
        variable_registry = get_variable_registry()
        CVS = variable_registry["complex_variable_systems"]
        if all(
            var in variable_registry["conversion_dictionaries"]["realToSym"]
            for var in varSpace
        ):
            _varSpace_type = "real"
        elif all(
            var in variable_registry["conversion_dictionaries"]["symToReal"]
            for var in varSpace
        ):
            _varSpace_type = "complex"

        exhaust1 = list(varSpace)
        populate = {
            "holVarDict": dict(),
            "antiholVarDict": dict(),
            "realVarDict": dict(),
            "imVarDict": dict(),
        }
        if _varSpace_type == "real":
            for var in varSpace:
                varStr = str(var)
                if var in exhaust1:
                    for parent in CVS.values():
                        if varStr in parent["variable_relatives"]:
                            cousin = (
                                set(
                                    parent["variable_relatives"][varStr][
                                        "complex_family"
                                    ][2:]
                                )
                                - {var}
                            ).pop()
                            if cousin in exhaust1:
                                exhaust1.remove(cousin)
                            if (
                                parent["variable_relatives"][varStr][
                                    "complex_positioning"
                                ]
                                == "real"
                            ):
                                realVar = var
                                exhaust1.remove(var)
                                imVar = cousin
                            else:
                                realVar = cousin
                                exhaust1.remove(var)
                                imVar = var
                            holVar = parent["variable_relatives"][varStr][
                                "complex_family"
                            ][0]
                            antiholVar = parent["variable_relatives"][varStr][
                                "complex_family"
                            ][1]
                            populate["holVarDict"][holVar] = [realVar, imVar]
                            populate["antiholVarDict"][antiholVar] = [realVar, imVar]
                            populate["realVarDict"][realVar] = [holVar, antiholVar]
                            populate["imVarDict"][imVar] = [holVar, antiholVar]
        else:  # self._varSpace_type == 'complex'
            for var in varSpace:
                varStr = str(var)
                if var in exhaust1:
                    for parent in CVS.values():
                        if varStr in parent["variable_relatives"]:
                            cousin = (
                                set(
                                    parent["variable_relatives"][varStr][
                                        "complex_family"
                                    ][:2]
                                )
                                - {var}
                            ).pop()
                            if cousin in exhaust1:
                                exhaust1.remove(cousin)
                            if (
                                parent["variable_relatives"][varStr][
                                    "complex_positioning"
                                ]
                                == "holomorphic"
                            ):
                                holVar = var
                                exhaust1.remove(var)
                                antiholVar = cousin
                            else:
                                holVar = cousin
                                exhaust1.remove(var)
                                antiholVar = var
                            realVar = parent["variable_relatives"][varStr][
                                "complex_family"
                            ][2]
                            imVar = parent["variable_relatives"][varStr][
                                "complex_family"
                            ][3]
                            populate["holVarDict"][holVar] = [realVar, imVar]
                            populate["antiholVarDict"][antiholVar] = [realVar, imVar]
                            populate["realVarDict"][realVar] = [holVar, antiholVar]
                            populate["imVarDict"][imVar] = [holVar, antiholVar]
        _realVarSpace = tuple(populate["realVarDict"].keys())
        _holVarSpace = tuple(populate["holVarDict"].keys())
        _antiholVarSpace = tuple(populate["antiholVarDict"].keys())
        _imVarSpace = tuple(populate["imVarDict"].keys())

        return [_realVarSpace + _imVarSpace, _holVarSpace, _antiholVarSpace]

    @property
    def metric(self):
        if self._metric is None:
            if not self.is_closed:
                warnings.warn(
                    "The provided symplectic form does not define a Kahler structure, so the associated metric tensor may not actually describe a metric."
                )

            coeffData = {
                (j, k): self.kahlerForm(
                    self.coor_frame[j], complex_struct_op(self.coor_frame[k])
                )
                for j in range(len(self.coor_frame))
                for k in range(j, len(self.coor_frame))
            }
            self._metric = metricClass(
                STFClass(self.varSpace, coeffData, 2, DGCVType="complex")
            )
        return self._metric

    @property
    def holRiemann(self):
        """
        The holomorphic Riemann curvature tensor, defined by hooking the complex structure operator into the second and fourth positions of the Kahler structure's metric's (0,4)-type Riemann curvature tensor.
        """
        if self._holRiemann is None:

            dim = len(self.coor_frame)
            R = self.metric.RiemannCurvature

            VFBasis = self.coor_frame

            def entry_rule(a, b, c, d):
                return sp.simplify(R(a, complex_struct_op(b), c, complex_struct_op(d)))

            coeffData = {
                (j, k, L, m): entry_rule(VFBasis[j], VFBasis[k], VFBasis[L], VFBasis[m])
                for j in range(dim)
                for k in range(j, dim)
                for L in range(dim)
                for m in range(dim)
            }
            self._holRiemann = tensorField(self.varSpace, coeffData, valence=(0,0,0,0), DGCVType="complex")
        return self._holRiemann

    @property
    def holRicci(self):
        """
        The holomorphic Ricci curvature tensor, defined by hooking the complex structure operator into the second position of the Kahler structure's metric's Ricci curvature tensor.
        """
        if self._holRicci is None:

            dim = len(self.coor_frame)
            Ric = self.metric.RicciTensor

            VFBasis = self.coor_frame

            def entry_rule(a, b):
                return sp.simplify(Ric(a, complex_struct_op(b)))

            coeffData = {
                (j, k): entry_rule(VFBasis[j], VFBasis[k])
                for j in range(dim)
                for k in range(j, dim)
            }
            self._holRicci = tensorField(self.varSpace, coeffData, valence=(0,0), DGCVType="complex")
        return self._holRicci

    @property
    def Bochner(self):
        r"""
        The Bochner curvature tensor in $T^{1,0}M\otimes T^{0,1}M\otimes T^{1,0}M\otimes T^{0,1}M$.
        """
        if self._Bochner is None:

            VFBasis = self.coor_frame_complex
            dim = len(VFBasis)
            compDim = int(sp.Rational(dim, 2))

            g = self.metric.SymTensorField
            R = self.metric.RiemannCurvature
            Ric = self.metric.RicciTensor
            S = self.metric.scalarCurvature

            def entry_rule(j, h, L, k):
                term1 = R(j, h, L, k)
                term2 = sp.Rational(1, compDim + 2) * (
                    (g(j, k)) * (Ric(L, h))
                    + (g(L, k)) * (Ric(j, h))
                    + (g(L, h)) * (Ric(j, k))
                    + (g(j, h)) * (Ric(L, k))
                )
                term3 = (
                    S
                    * sp.Rational(1, 2 * (compDim + 1) * (compDim + 2))
                    * (g(L, h) * g(j, k) + g(j, h) * g(L, k))
                )
                return sp.simplify(term1 + term2 - term3)

            coeffData = {
                (j, k, L, m): entry_rule(VFBasis[j], VFBasis[k], VFBasis[L], VFBasis[m])
                for j in range(compDim)
                for k in range(compDim, dim)
                for L in range(compDim)
                for m in range(compDim, dim)
            }
            self._Bochner = tensorField(
                self.varSpace,
                coeffData,
                valence=(0,0,0,0),
                DGCVType="complex",
                _simplifyKW={
                    "simplify_rule": None,
                    "simplify_ignore_list": None,
                    "preferred_basis_element": (0, compDim, 0, compDim),
                },
            )

        return self._Bochner

    @property
    def is_closed(self):
        if self._is_closed is None:
            self._is_closed = exteriorDerivative(self.kahlerForm).is_zero()
        return self._is_closed




from .._safeguards import _cached_caller_globals, create_key, retrieve_passkey
from ..dgcv_core import clearVar, listVar, variableProcedure
from .eds import abst_coframe, abstDFAtom, abstDFMonom, abstract_DF, extDer


def transform_coframe(original_coframe:abst_coframe, transformations:list|tuple|dict, new_coframe_basis = None, new_coframe_labels = None, min_conj_rules = {}):
    original_basis = original_coframe.forms
    if new_coframe_basis is None:
        if isinstance(new_coframe_labels, (list,tuple)) and all(isinstance(term,str) for term in new_coframe_labels) and len(new_coframe_labels) == len(original_basis):
            new_coframe_basis = [abstDFAtom(1,1,label=term) for term in new_coframe_labels]
        else:
            raise ValueError('`transform_coframe` must either be given a list of DF objects for `new_coframe_basis` or a list of string labels for `new_coframe_labels` from which to create a new basis.')
    elif len(new_coframe_basis)!=len(original_basis):
        raise ValueError(f'If provided, the `new_coframe_basis` list given to `transform_coframe` must match the original coframe\'s basis size. In this case the list had length {len(new_coframe_basis)} while {len(original_basis)} was expected.')

    if isinstance (transformations,(list,tuple)) and all(isinstance(term,(list,tuple)) for term in transformations) and all(len(term)==2 for term in transformations):
        transformations = {term[0]:term[1] for term in transformations}
    if isinstance(transformations,dict):
        new_basis = tuple([df.subs(transformations) for df in original_basis])
    elif isinstance(transformations,(list,tuple)) and all(isinstance(term,(abstract_DF,abstDFMonom,abstDFAtom)) for term in transformations) and len(transformations)==len(original_basis) and min_conj_rules=={}:
        new_basis = transformations
    else:
        if len(min_conj_rules) == {}:
            raise TypeError(f'transformation rules given to `transform_coframe` must be either a dictionary (or dict-like list of pairs) whose keys belong to the old coframe basis and whose values are the new DF they transform to, or a list of new DF with proper length expressed in terms of the old basis: not {transformations}')
        else:
            raise TypeError(f'If providing `min_conj_rules`, transformation rules given to `transform_coframe` must be a dictionary (or dict-like list of pairs) whose keys belong to the old coframe basis and whose values are the new DF they transform to expressed in terms of the old basis: not {transformations}')
    new_basis_squares = []
    new_basis_squares_labels = []
    for j, el1 in enumerate(new_basis):
        el1Alt = new_coframe_basis[j]
        for k, el2 in enumerate(new_basis[j+1:]):
            el2Alt = new_coframe_basis[j+1:][k]
            new_basis_squares += [el1*el2]
            new_basis_squares_labels += [el1Alt*el2Alt]
    varLabel=create_key(prefix="coeff")             # label for temparary variables
    variableProcedure(
        varLabel,
        len(new_basis_squares),
        _tempVar=retrieve_passkey()
    )
    cVars = _cached_caller_globals[varLabel]        # pointer to tuple of coef vars

    general_elem = sum([j*k for j,k in zip(cVars[1:],new_basis_squares[1:])],cVars[0]*new_basis_squares[0])
    general_elem_new_basis = sum([j*k for j,k in zip(cVars[1:],new_basis_squares_labels[1:])],cVars[0]*new_basis_squares_labels[0])
    new_structure_eqns = dict()

    for df_atom, df in zip(new_coframe_basis,new_basis):
        from ..solvers import solve_DGCV
        new_extD = extDer(df,original_coframe)
        eqns = [new_extD - general_elem]
        solution = solve_DGCV(eqns,cVars)
        if len(solution)>1:
            raise ValueError('The given coframe transformation rule is not invertible')
        if len(solution)<1:
            raise ValueError('Unable to compute new coframe structure equations w.r.t. given transformation rule')
        new_structure_eqns |= {df_atom:general_elem_new_basis.subs(solution[0])}
    clearVar(*listVar(temporary_only=True), report=False)

    return abst_coframe(new_coframe_basis,new_structure_eqns,min_conj_rules=min_conj_rules)



from sympy import Basic, im, re, simplify

from .complex_structures import KahlerStructure
from .config import get_variable_registry
from .dgcv_core import DFClass, STFClass, TFClass, VF_coeffs, addVF, allToReal, allToSym
from .Riemannian_geometry import metricClass
from .vector_fields_and_differential_forms import get_DF, get_VF


class coordinate_map(Basic):
    def __new__(cls, coordinates1, coordinates2, coordinate_formulas, holomorphic=None):
        obj = Basic.__new__(cls, coordinates1, coordinates2, coordinate_formulas)
        return obj

    def __init__(
        self, coordinates1, coordinates2, coordinate_formulas, holomorphic=None
    ):
        if not all(
            isinstance(j, (list, tuple))
            for j in [coordinates1, coordinates2, coordinate_formulas]
        ):
            raise TypeError(
                "`coordinate_map` needs all arguments given in the form of lists or tuples."
            )
        if len(coordinates2) != len(coordinate_formulas):
            raise TypeError(
                "`coordinate_map` recieved incompatible initialization data. The number of coordinates provided for the range (i.e., in the second argument position) must match the number of components in the provided formulas for the mapping's components (i.e., in the third argument position)."
            )

        vr = get_variable_registry()

        self.domain = tuple(coordinates1)
        self.range = tuple(coordinates2)
        if holomorphic:

            def get_real_parts(varList):
                reals = []
                ims = []
                for var in varList:
                    varStr = str(var)
                    for parent in vr["complex_variable_systems"]:
                        if (
                            varStr
                            in vr["complex_variable_systems"][parent][
                                "variable_relatives"
                            ]
                        ):
                            realParts = vr["complex_variable_systems"][parent][
                                "variable_relatives"
                            ][varStr]["complex_family"][2:]
                            reals = reals + [realParts[0]]
                            ims = ims + [realParts[1]]
                return reals + ims

            if all(
                j in vr["conversion_dictionaries"]["symToReal"]
                for j in self.domain + self.range
            ):
                self.domain = get_real_parts(self.domain)
                self.range = get_real_parts(self.range)
                coordinate_formulas = [
                    simplify(re(allToReal(j))) for j in coordinate_formulas
                ] + [simplify(im(allToReal(j))) for j in coordinate_formulas]
            else:
                raise TypeError(
                    "When setting `holomorphic=True`, `coordinate_map` expects the variables given for the domain and range to be all holomorphic parts of DGCV complex variable systems. It will infer the appropriate action of antiholomophic/real/imaginary parts using holomorphicity."
                )
        self.domain_varSpace_type, self.domain_frame, self.domain_coframe = (
            coordinate_map.validate_coordinates(self.domain)
        )
        self.range_varSpace_type, self.range_frame, self.range_coframe = (
            coordinate_map.validate_coordinates(self.range)
        )
        self._varSpace_type = self.domain_varSpace_type
        self.coordinate_formulas = list(coordinate_formulas)
        self._JacobianMatrix = None

    @property
    def JacobianMatrix(self):
        if self._JacobianMatrix is None:
            self._JacobianMatrix = [
                [simplify(j(k)) for j in self.domain_frame]
                for k in self.coordinate_formulas
            ]
        return self._JacobianMatrix

    @staticmethod
    def validate_coordinates(varSpace):

        if len(varSpace) != len(set(varSpace)):
            raise TypeError(
                "`coordinate_map` was a list of variables for coordinates (either for the domain or range) that have repeated values.)"
            )
        vr = get_variable_registry()

        def checkReal(var):
            return var in vr["conversion_dictionaries"]["realToSym"]

        def checkHol(var):
            return var in vr["conversion_dictionaries"]["symToReal"]

        def checkStandard(var):
            varStr = str(var)
            for parent in vr["standard_variable_systems"]:
                if (
                    varStr
                    in vr["standard_variable_systems"][parent]["variable_relatives"]
                ):
                    return True
            return False

        def checkRegistered(var):
            varStr = str(var)
            for parent in vr["complex_variable_systems"]:
                if (
                    varStr
                    in vr["complex_variable_systems"][parent]["variable_relatives"]
                ):
                    return True
            return checkStandard(var)

        if any(checkReal(var) for var in varSpace):
            if all(checkReal(var) for var in varSpace):
                return (
                    "real",
                    [allToReal(j) for j in get_VF(*varSpace)],
                    [allToReal(j) for j in get_DF(*varSpace)],
                )
            else:
                raise TypeError(
                    "`coordinate_map` was given coordinates (either for the domain or range) that are all associated with complex variable systems but given with non-uniform variable format between 'real coordinates' (i.e., real and imaginary parts of holomorphic coordinates) and 'holomorphic coordinates'. Coordinates should be provided in a uniform format, be it real or holomorphic. (The format may vary between domain and range.)"
                )
        elif all(checkHol(var) for var in varSpace):
            return (
                "complex",
                [allToSym(j) for j in get_VF(*varSpace)],
                [allToSym(j) for j in get_DF(*varSpace)],
            )
        elif all(checkStandard(var) for var in varSpace):
            return "standard", get_VF(*varSpace), get_DF(*varSpace)
        elif not all(checkRegistered(var) for var in varSpace):
            raise TypeError(
                "`coordinate_map` was given coordinates containing variables that were not initialized in the DGCV variable management framework (VMF). Use variable creation functions like `createVariables` to initialize variables while automatically registering them in the VMF."
            )
        else:

            raise TypeError(
                "`coordinate_map` was given coordinates (either for the domain or range) that contain a mixture of variable types, i.e., some from a 'standard' variable system and some from a 'complex' variable system. Only one type is allowed per coordinate set (but the type can vary between domain and range)."
            )

    def differential(self, vf):
        inputCoeffs = VF_coeffs(vf, self.domain)
        vf_list = []
        for j in range(len(self.range)):
            vf_list += [
                sum(
                    [
                        inputCoeffs[k] * self.JacobianMatrix[j][k]
                        for k in range(len(self.domain))
                    ]
                )
                * self.range_frame[j]
            ]
        return addVF(*vf_list)

    def pull_back(self, tf):
        if isinstance(tf, STFClass):
            vf_basis = [self.differential(vf) for vf in self.domain_frame]
            if set(tf.varSpace) == set(self.range):

                def entry_rule(
                    iT,
                ):  # define tensor field value for a particular index tuple
                    argumentList = [vf_basis[iT[j]] for j in range(len(iT))]
                    return tf(*argumentList)

                deg = tf.degree

                def generate_nondecreasing_indices(shape, min_val=0):
                    """
                    Recursively generates all index tuples of nondecreasing integer lists for an arbitrary dimensional array.

                    Parameters
                    ----------
                    shape : tuple
                        A tuple where each element defines the dimension (length) of each axis.
                    min_val : int, optional
                        The minimum value allowed for the current index, ensuring nondecreasing order.

                    Returns
                    -------
                    list of tuples
                        All possible index tuples with nondecreasing integer values.
                    """
                    if len(shape) == 1:
                        # Generate the indices for the last dimension, constrained by min_val
                        return [(i,) for i in range(min_val, shape[0])]
                    else:
                        # Recursively generate indices for the next dimensions, ensuring nondecreasing order
                        return [
                            (i,) + t
                            for i in range(min_val, shape[0])
                            for t in generate_nondecreasing_indices(shape[1:], i)
                        ]

                shape = (len(self.domain),) * deg
                sparse_data = {
                    indices: entry_rule(indices)
                    for indices in generate_nondecreasing_indices(shape)
                }
                return STFClass(
                    self.domain,
                    sparse_data,
                    deg,
                    DGCVType=tf.DGCVType,
                    _simplifyKW=tf._simplifyKW,
                )

            else:
                raise Exception(
                    "`coordinate_map.pull_back` can only be applied to differential for tensor fields defined over the same coordinates initialized as `coodinate_map.range`."
                )
        if isinstance(tf, TFClass):
            vf_basis = [self.differential(vf) for vf in self.domain_frame]
            if set(tf.varSpace) == set(self.range):

                def entry_rule(
                    iT,
                ):  # define tensor field value for a particular index tuple
                    argumentList = [vf_basis[iT[j]] for j in range(len(iT))]
                    return tf(*argumentList)

                deg = tf.degree

                def generate_indices(shape):
                    """Recursively generates all index tuples for an arbitrary dimensional array."""
                    if len(shape) == 1:
                        return [(i,) for i in range(shape[0])]
                    else:
                        return [
                            (i,) + t
                            for i in range(shape[0])
                            for t in generate_indices(shape[1:])
                        ]

                dim = len(self.domain)
                shape = (dim,) * 3
                sparse_data = {
                    indices: entry_rule(indices) for indices in generate_indices(shape)
                }

                shape = (len(self.domain),) * deg
                sparse_data = {
                    indices: entry_rule(indices) for indices in generate_indices(shape)
                }
                return TFClass(
                    self.domain,
                    sparse_data,
                    deg,
                    DGCVType=tf.DGCVType,
                    _simplifyKW=tf._simplifyKW,
                )

            else:
                raise Exception(
                    "`coordinate_map.pullack` can only be applied to differential for tensor fields defined over the same coordinates initialized as `coodinate_map.range`."
                )
        if isinstance(tf, DFClass):
            vf_basis = [self.differential(vf) for vf in self.domain_frame]
            if set(tf.varSpace) == set(self.range):

                def entry_rule(
                    iT,
                ):  # define tensor field value for a particular index tuple
                    argumentList = [vf_basis[iT[j]] for j in range(len(iT))]
                    return tf(*argumentList)

                deg = tf.degree

                def generate_nondecreasing_indices(shape, min_val=0):
                    """
                    Recursively generates all index tuples of increasing integer lists for an arbitrary dimensional array.

                    Parameters
                    ----------
                    shape : tuple
                        A tuple where each element defines the dimension (length) of each axis.
                    min_val : int, optional
                        The minimum value allowed for the current index, ensuring nondecreasing order.

                    Returns
                    -------
                    list of tuples
                        All possible index tuples with strictly increasing integer values.
                    """
                    if len(shape) == 1:
                        # Generate the indices for the last dimension, constrained by min_val
                        return [(i,) for i in range(min_val, shape[0])]
                    else:
                        # Recursively generate indices for the next dimensions, ensuring strictly increasing order
                        return [
                            (i,) + t
                            for i in range(min_val, shape[0])
                            for t in generate_nondecreasing_indices(shape[1:], i + 1)
                        ]

                shape = (len(self.domain),) * deg
                sparse_data = {
                    indices: entry_rule(indices)
                    for indices in generate_nondecreasing_indices(shape)
                }
                return DFClass(
                    self.domain,
                    sparse_data,
                    deg,
                    DGCVType=tf.DGCVType,
                    _simplifyKW=tf._simplifyKW,
                )

            else:
                raise Exception(
                    "`coordinate_map.pullack` can only be applied to differential for tensor fields defined over the same coordinates initialized as `coodinate_map.range`."
                )
        if isinstance(tf, metricClass):
            return metricClass(self.pull_back(tf.SymTensorField))

        if isinstance(tf, KahlerStructure):
            newKF = self.pull_back(tf.kahlerForm)
            newVS = newKF.varSpace
            return KahlerStructure(newVS, newKF)
